import datetime
import os
from functools import partial

import numpy as np
import requests

from pywapor.collect.protocol import cds
from pywapor.enhancers.pressure import pa_to_kpa
from pywapor.enhancers.temperature import kelvin_to_celsius
from pywapor.enhancers.wind import adjust_wind_height, windspeed
from pywapor.general.logger import adjust_logger
from pywapor.general.processing_functions import open_ds, remove_ds


def default_vars(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary
    with metadata on which exact layers need to be requested from the server, how they should
    be renamed, and how their dimensions are defined.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Metadata on which exact layers need to be requested from the server.
    """

    variables = {
        "sis-agrometeorological-indicators": {
            "2m_temperature": [
                {"statistic": "24_hour_mean", "format": "zip", "version": "1_1"},
                "t_air",
            ],
            "2m_dewpoint_temperature": [
                {"statistic": "24_hour_mean", "format": "zip", "version": "1_1"},
                "t_dew",
            ],
            "2m_relative_humidity": [
                {
                    "time": [f"{x:02d}_00" for x in [6, 9, 12, 15, 18]],
                    "format": "zip",
                    "version": "1_1",
                },
                "rh",
            ],
            "10m_wind_speed": [
                {"statistic": "24_hour_mean", "format": "zip", "version": "1_1"},
                "u",
            ],
            "vapour_pressure": [
                {"statistic": "24_hour_mean", "format": "zip", "version": "1_1"},
                "vp",
            ],
            "solar_radiation_flux": [{"format": "zip", "version": "1_1"}, "ra_flat"],
        },
        "reanalysis-era5-single-levels": {
            "total_column_water_vapour": [
                {
                    "time": [f"{x:02d}:00" for x in range(24)],
                    "format": "netcdf",
                    "product_type": "reanalysis",
                },
                "wv",
            ],
            "10m_u_component_of_wind": [
                {
                    "time": [f"{x:02d}:00" for x in range(24)],
                    "format": "netcdf",
                    "product_type": "reanalysis",
                },
                "u10m",
            ],
            "10m_v_component_of_wind": [
                {
                    "time": [f"{x:02d}:00" for x in range(24)],
                    "format": "netcdf",
                    "product_type": "reanalysis",
                },
                "v10m",
            ],
            "2m_dewpoint_temperature": [
                {
                    "time": [f"{x:02d}:00" for x in range(24)],
                    "format": "netcdf",
                    "product_type": "reanalysis",
                },
                "t_dew",
            ],
            "mean_sea_level_pressure": [
                {
                    "time": [f"{x:02d}:00" for x in range(24)],
                    "format": "netcdf",
                    "product_type": "reanalysis",
                },
                "p_air_0",
            ],
            "surface_pressure": [
                {
                    "time": [f"{x:02d}:00" for x in range(24)],
                    "format": "netcdf",
                    "product_type": "reanalysis",
                },
                "p_air",
            ],
            "2m_temperature": [
                {
                    "time": [f"{x:02d}:00" for x in range(24)],
                    "format": "netcdf",
                    "product_type": "reanalysis",
                },
                "t_air",
            ],
        },
    }

    req_dl_vars = {
        "sis-agrometeorological-indicators": {
            "t_air": ["2m_temperature"],
            "t_air_max": ["2m_temperature"],
            "t_air_min": ["2m_temperature"],
            "t_dew": ["2m_dewpoint_temperature"],
            "rh": ["2m_relative_humidity"],
            "u": ["10m_wind_speed"],
            "vp": ["vapour_pressure"],
            "ra_flat": [
                "solar_radiation_flux"
            ],  # [J m-2d-1], Downward solar radiation ('Solar_Radiation_Flux')
        },
        "reanalysis-era5-single-levels": {
            "wv": ["total_column_water_vapour"],
            "u10m": ["10m_u_component_of_wind"],
            "v10m": ["10m_v_component_of_wind"],
            "u2m": ["10m_u_component_of_wind"],
            "v2m": ["10m_v_component_of_wind"],
            "u": ["10m_u_component_of_wind", "10m_v_component_of_wind"],
            "t_dew": ["2m_dewpoint_temperature"],
            "p_air_0": ["mean_sea_level_pressure"],
            "p_air": ["surface_pressure"],
            "t_air": ["2m_temperature"],
            "t_air_min": ["2m_temperature"],
            "t_air_max": ["2m_temperature"],
        },
    }

    out = {
        val: variables[product_name][val]
        for sublist in map(req_dl_vars[product_name].get, req_vars)
        for val in sublist
    }

    return out


def jouleperday_to_watt(ds, var):
    ds[var] = ds[var] / 86400
    return ds


def default_post_processors(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary with a
    list of functions per variable that should be applied after having collected the data
    from a server.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Functions per variable that should be applied to the variable.
    """

    post_processors = {
        "sis-agrometeorological-indicators": {
            "t_air": [kelvin_to_celsius],
            "t_air_max": [
                partial(kelvin_to_celsius, in_var="t_air", out_var="t_air_max")
            ],
            "t_air_min": [
                partial(kelvin_to_celsius, in_var="t_air", out_var="t_air_min")
            ],
            "t_dew": [kelvin_to_celsius],
            "rh": [],
            "u": [adjust_wind_height],
            "vp": [],  # is already mbar
            "ra_flat": [jouleperday_to_watt],
        },
        "reanalysis-era5-single-levels": {
            "wv": [],  # is already kg/m2
            "u10m": [],
            "v10m": [],
            "u2m": [adjust_wind_height],
            "v2m": [adjust_wind_height],
            "u": [windspeed, adjust_wind_height],
            "t_dew": [kelvin_to_celsius],
            "p_air_0": [pa_to_kpa],
            "p_air": [pa_to_kpa],
            "t_air": [kelvin_to_celsius],
            "t_air_max": [
                partial(kelvin_to_celsius, in_var="t_air", out_var="t_air_max")
            ],
            "t_air_min": [
                partial(kelvin_to_celsius, in_var="t_air", out_var="t_air_min")
            ],
        },
    }

    out = {k: v for k, v in post_processors[product_name].items() if k in req_vars}

    return out


def most_recent(product_name, *args):
    url = f"https://cds.climate.copernicus.eu/api/catalogue/v1/collections/{product_name}"
    r = requests.get(url)
    time_interval = r.json()["extent"]["temporal"]["interval"][0]
    end_time = time_interval[1]
    return datetime.datetime.strptime(end_time, "%Y-%m-%dT%H:%M:%SZ")


def download(
    folder,
    latlim,
    lonlim,
    timelim,
    product_name,
    req_vars,
    variables=None,
    post_processors=None,
):
    """Download ERA5 data and store it in a single netCDF file.

    Parameters
    ----------
    folder : str
        Path to folder in which to store results.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    product_name : str
        Name of the product to download.
    req_vars : list
        Which variables to download for the selected product.
    variables : dict, optional
        Metadata on which exact layers need to be requested from the server, by default None.
    post_processors : dict, optional
        Functions per variable that should be applied to the variable, by default None.

    Returns
    -------
    xr.Dataset
        Downloaded data.
    """
    product_folder = os.path.join(folder, "ERA5")

    if not os.path.exists(product_folder):
        os.makedirs(product_folder)

    fn_final = os.path.join(product_folder, f"{product_name}.nc")
    if os.path.isfile(fn_final):
        ds = open_ds(fn_final)
        if np.all([x in ds.data_vars for x in req_vars]):
            return ds[req_vars]
        else:
            remove_ds(ds)

    spatial_buffer = True
    if spatial_buffer:
        buffer = {
            "sis-agrometeorological-indicators": 0.1,
            "reanalysis-era5-single-levels": 0.25,
        }[product_name]
        latlim = [latlim[0] - buffer, latlim[1] + buffer]
        lonlim = [lonlim[0] - buffer, lonlim[1] + buffer]

    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {
            k: {True: default_processors[k], False: v}[v == "default"]
            for k, v in post_processors.items()
        }

    ds = cds.download(
        product_folder,
        product_name,
        latlim,
        lonlim,
        timelim,
        variables,
        post_processors,
    )

    return ds


if __name__ == "__main__":
    folder = r"/Users/hmcoerver/Local/era_test"
    latlim = [28.9, 29.7]
    lonlim = [30.2, 31.2]
    timelim = ["2021-06-26", "2021-07-05"]

    # product_name = "sis-agrometeorological-indicators"
    product_name = "reanalysis-era5-single-levels"
    req_vars = ["t_air", "t_dew"]  # , "vp", "ra"]
    # req_vars = ["u"]

    variables = None
    post_processors = None

    adjust_logger(True, folder, "INFO")

    ds = download(
        folder,
        latlim,
        lonlim,
        timelim,
        product_name=product_name,
        req_vars=req_vars,
        variables=variables,
        post_processors=post_processors,
    )



"""
    The biomass module contains all functions related to biomass production
    data components.

"""
import numpy as np
import xarray as xr

def fpar(ndvi, fpar_slope=1.257, fpar_offset=-0.161): # TODO: UPDATE WITH METHODOLOGY
    r"""
    Computes the fraction of absorbed PAR.
    PAR is photosynthetically active radiation, which is the solar radiation
    in the spectral range from 400-700 nm that is used by plants in
    photosynthesis. This function computes the fpar as a linear function of
    the ndvi, and is bounded between 0 and 1.

    .. math ::
        f_{par}=\Delta_{p} \cdot I_{ndvi} + c_{p}

    Parameters
    ----------
    ndvi : float
        normalized difference vegetation index
        :math:`I_{ndvi}`
        [-]

    fpar_slope : float
        slope of the fpar-ndvi curve
        :math:`\Delta_{p}`
        [-]

    fpar_offset : float
        offset of the fpar-ndvi curve
        :math:`c_{p}`
        [-]

    Returns
    -------
    f_par : float
        fraction of absorbed photosynthetically active radiation
        :math:`f_{par}`
        [-]

    Examples
    --------
    >>> import qattara.components.biomass as bio
    >>> bio.fpar(0.4)
    0.34179999999999999

    """
    x = fpar_offset + fpar_slope * ndvi

    if isinstance(x, xr.DataArray):
        out = x.clip(0.0,1.0)
    else:
        out = np.clip(x, 0.0, 1.0)
    return out


def par(ra_24):
    r"""
    Computes the photosynthetically active radiation (PAR). PAR is the solar
    radiation in the spectral range from 400-700 nm that is used by plants
    for photosynthesis.

    .. math ::
        PAR=r_{PAR} \cdot S^{\downarrow}

    where the following constant is used

    * :math:`r_{PAR}` = ratio par/solar radiation [-] = 0.48

    Parameters
    ----------
    ra_24 : float
       incoming shortwave radiation
       :math:`S^{\downarrow}`
       [W/m2]

    Returns
    -------
    apar : float
       photosynthetically active radiation
       :math:`PAR`
       [W/m2]

    Examples
    --------
    >>> import qattara.components.biomass as bio
    >>> bio.par(400.0)
    192.0

    """
    ABSORBED_RADIATION = 0.48

    return ABSORBED_RADIATION * ra_24

def co2_level_annual(year):
    r"""
    Computes the annual CO2 level based on the linear relation of Roel Van Hoolst (Copernicus).

    Parameters
    ----------
    year : integer
        year of interest
        :math:`y`
        [-]

    Returns
    -------
    co2_act : float
        actual annual CO2 level
        :math:`CO_2act`
        [ppmv]

    Examples
    --------
    #>>> from qattara import biomass
    #>>> import numpy as np
    #>>> biomass.t_air_kelvin(np.array([12.5,18.3]))
    #array([ 285.65,  291.45])
    """
    return 2.0775 * year - 3785.783


def temperature_dependency(t_air_k_12, dh_ap=52750, d_s=704.98, dh_dp=211000):
    r"""
    Computes the temperature dependency factor of GPP.

    Parameters
    ----------
    t_air_k_12 : float
        daytime air temperature
        :math:`T_{a,12}`
        [K]

    dh_ap : float
        activation energy
        :math:`\delta DH_{a,P}`
        [J/mol]

    d_s : float
        entropy of the denaturation equilibrium of CO2
        :math:`\delta S`
        [J/K.mol]

    dh_dp : float
        deactivation energy
        :math:`\delta H_{d,P}`
        [J/mol]

    Returns
    -------
    t_dep : float
        temperature dependency factor
        :math:`p(T_{atm})`
        [-]

    """
    R_G = 8.3144 # The molar gas constant in J K -1 mol-1

    c1 = 21.77
    x = R_G * t_air_k_12
    y = c1 - dh_ap / x
    z = (d_s * t_air_k_12 - dh_dp) / x
    return np.exp(y) / (1 + np.exp(z))


def co2_o2_specificity_ratio(t_air_k_12):
    r"""
    Computes the CO2/O2 specificity ratio

    Parameters
    ----------
    t_air_k_12 : float
        daytime air temperature
        :math:`T_{a,12}`
        [K]

    Returns
    -------
    tau_co2_o2 : float
        CO2/O2 specificity ratio
        :math:`\tau_CO{2}O{2}`
        [-]
    """
    R_G = 8.3144 # The molar gas constant in J K -1 mol-1

    a_t = 7.87 * 10 ** -5
    e_t = -42869.9
    x = R_G * t_air_k_12
    return a_t * np.exp(-e_t / x)


def inhibition_constant_o2(t_air_k_12):
    r"""
    Computes the inhibition constant for O2

    Parameters
    ----------
    t_air_k_12 : float
        daytime air temperature
        :math:`T_{a,12}`
        [K]

    Returns
    -------
    k_0 : float
        inhibition constant for O2
        :math:`K_0`
        [% O2]
    """
    R_G = 8.3144 # The molar gas constant in J K -1 mol-1

    a_0 = 8240
    e_0 = 13913.5
    x = R_G * t_air_k_12
    return a_0 * np.exp(-e_0 / x)


def affinity_constant_co2(t_air_k_12):
    r"""
    Computes the affinity constant for CO2 of Rubisco

    Parameters
    ----------
    t_air_k_12 : float
        daytime air temperature
        :math:`T_{a,12}`
        [K]

    Returns
    -------
    k_m : float
        affinity constant for CO2 of Rubisco
        :math:`K_m`
        [% CO2]
    """
    a1 = 2.419 * 10 ** 13
    a2 = 1.976 * 10 ** 22
    e1 = 59400
    e2 = 109600

    R_G = 8.3144 # The molar gas constant in J K -1 mol-1
    
    if isinstance(t_air_k_12, xr.DataArray):
        x1 = a1 * np.exp(-e1 / (R_G * t_air_k_12))
        x2 = a2 * np.exp(-e2 / (R_G * t_air_k_12))
        k_m = xr.where(t_air_k_12 >= 288.13, x1, x2)
    else:
        k_m = np.zeros(t_air_k_12.shape)
        k_m[t_air_k_12 >= 288.13] = a1 * np.exp(-e1 / (R_G * t_air_k_12[t_air_k_12 >= 288.13]))
        k_m[t_air_k_12 < 288.13] = a2 * np.exp(-e2 / (R_G * t_air_k_12[t_air_k_12 < 288.13]))
    return k_m


def co2_fertilisation(tau_co2_o2, k_m, k_0, co2_act, o2=20.9, co2_ref=281):
    r"""
    Computes the normalized CO2 fertilization factor (Veroustraete, 1994). No fertilization means values equal to 1. Fertilization means values larger than 1.

    Parameters
    ----------
    tau_co2_o2 : float
        CO2/O2 specificity ratio
        :math:`\tau_CO{2}O{2}`
        [-]

    k_m : float
        affinity constant for CO2 of Rubisco
        :math:`K_m`
        [% CO2]

    k_0 : float
        inhibition constant for O2
        :math:`K_0`
        [% O2]

    co2_act : float
        actual annual CO2 level
        :math:`CO_2act`
        [ppmv]

    o2 : float
        O2 concentration
        :math:`O_2`
        [%]

    co2_ref : float
        reference CO2 level
        :math:`CO_2ref`
        [ppmv]

    Returns
    -------
    co2_fert : float
        CO2 fertilization effect
        :math:`CO_2fert`
        [-]

    """
    Y = o2 / (2 * tau_co2_o2)
    Z = k_m * (1 + o2 / k_0)
    return (co2_act - Y) / (co2_ref - Y) * (Z + co2_ref) / (Z + co2_act)


def autotrophic_respiration(t_air_k_24, ar_slo=0.0, ar_int=0.5):
    r"""
    Computes the fraction lost to autotrophic respiration

    Parameters
    ----------
    t_air_k_24 : float
        daily air temperature
        :math:`T_{a,24}`
        [K]

    ar_slo : float
        slope to determine autotrophic respiratory fraction
        :math:`AR_{slo}`
        [-]

    ar_int : float
        intercept to determine autotrophic respiratory fraction
        :math:`AR_{int}`
        [-]

    Returns
    -------
    a_d : float
         autotrophic respiratory fraction
        :math:`A_d`
        [-]
    """
    return ar_slo * t_air_k_24 + ar_int


def net_primary_production_max(t_dep, co2_fert, a_d, apar, gcgdm=0.45):
    r"""
    Computes the maximum Net Primary Production.

    Parameters
    ----------
    t_dep : float
        temperature dependency factor
        :math:`p(T_{atm})`
        [-]

    co2_fert : float
        CO2 fertilization effect
        :math:`CO_2fert`
        [-]

    a_d : float
        autotrophic respiratory fraction
        :math:`A_d`
        [-]

    apar : float
        photosynthetic active radiation
        :math:`PAR`
        [W/m2]

    gcgdm : float
        conversion factor enabling the conversion from DM to C
        :math:`gCgDM`
        [-]

    Returns
    -------
    npp_max : float
        maximum net primary production
        :math:`NPP_{max}`
        [gC/m2]

    """
    apar_megajoule_per_day = apar * (60 * 60 * 24) * 10 ** (-6)  # second to day, joule to megajoule
    return t_dep * co2_fert * apar_megajoule_per_day * gcgdm * (1 - a_d)


def net_primary_production(npp_max, f_par, stress_moist, phot_eff=2.49):
    r"""
    Computes the maximum Net primary Production.

    Parameters
    ----------
    npp_max : float
        maximum net primary production
        :math:`NPP_{max}`
        [-]

    phot_eff : float
        Default Radiation Use Efficiency
        :math:` \eps`
        [gDM/MJ(APAR)]

    f_par : float
        fraction of absorbed photosynthetically active radiation
        :math:`f_{par}`
        [-]

    stress_moist : float
        stress factor of root zone moisture
        :math:`S_{m}`
        [-]

    Returns
    -------
    npp : float
        net primary production
        :math:`NPP`
        [gC/m2]
    """
    return npp_max * phot_eff * f_par * stress_moist




"""
Generates input data for `pywapor.et_look`.
"""
from pywapor.collect import downloader
from pywapor.general.logger import log, adjust_logger
from pywapor.general import compositer
import pywapor.general.levels as levels
import datetime
import numpy as np
import os
import pandas as pd
import xarray as xr
from functools import partial
import pywapor.general.pre_defaults as defaults
from pywapor.general.variables import fill_attrs
from pywapor.enhancers.temperature import lapse_rate_to_all

def rename_vars(ds, *args):
    """Rename some variables in a dataset.

    Parameters
    ----------
    ds : xr.Dataset
        Input dataset whose variables will be renamed.

    Returns
    -------
    xr.Dataset
        Dataset with renamed variables.
    """
    varis = ["p", "ra_flat", "t_air", "t_air_min", "t_air_max", "u", "vp",
            "u2m", "v2m", "qv", "p_air", "p_air_0", "wv", "t_dew"]
    present_vars = [x for x in varis if x in ds.variables]
    ds = ds.rename({k: k + "_24" for k in present_vars})
    return ds

def calc_doys(ds, *args, bins = None):
    """Calculate the day-of-the-year (doy) in the middle of a timebin and assign the results to a new
    variable `doy`.

    Parameters
    ----------
    ds : xr.Dataset
        Input dataset.
    bins : list, optional
        List of boundaries of the timebins, by default None

    Returns
    -------
    xr.Dataset
        Dataset with a new variable containing the `doy` per timebin.
    """
    bin_doys = [int(pd.Timestamp(x).strftime("%j")) for x in bins]
    doy = np.mean([bin_doys[:-1], bin_doys[1:]], axis=0, dtype = int)
    if "time_bins" in list(ds.variables):
        ds["doy"] = xr.DataArray(doy, coords = ds["time_bins"].coords).chunk("auto")
    return ds

def add_constants(ds, *args):
    """Adds default dimensionless constants to a dataset.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to which the constants will be added.

    Returns
    -------
    xr.Dataset
        Dataset with extra variables.
    """
    ds = ds.assign(defaults.constants_defaults())
    return ds

def remove_empty_statics(ds, *args):
    for var in ds.data_vars:
        if "time" not in ds[var].coords and "time_bins" not in ds[var].coords:
            if ds[var].isnull().all():
                ds = ds.drop(var)
                log.info(f"--> Removing `{var}` from dataset since it is empty.")
    return ds

def add_constants_new(ds, *args): # TODO make sure this only adds constants for et_look, not for se_root.
    c = defaults.constants_defaults()
    for var, value in c.items():
        if var not in ds.data_vars:
            ds[var] = value
    return ds

def main(folder, latlim, lonlim, timelim, sources = "level_2_v3", bin_length = 1, enhancers = [lapse_rate_to_all]):
    """Prepare input data for `et_look`.

    Parameters
    ----------
    folder : str
        Path to folder in which to store (intermediate) data.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    sources : "level_1" | "level_2" | "level_2_v3" | dict, optional
        Configuration for each variable and source, by default `"level_1"`.
    bin_length : int | "DEKAD", optional
        Composite length, by_default 1.
    enhancers : list, optional
        Functions to apply to the xr.Dataset before creating the final
        output, by default `[lapse_rate]`.

    Returns
    -------
    xr.Dataset
        Dataset with input data for `pywapor.et_look`.
    """
    if not os.path.exists(folder):
        os.makedirs(folder)

    _ = adjust_logger(True, folder, "INFO")

    log.info(f"--> lonlim: {lonlim}")
    log.info(f"--> latlim: {latlim}")
    log.info(f"--> timelim: {timelim}")

    t1 = datetime.datetime.now()
    log.info("> PRE_ET_LOOK").add()

    if isinstance(sources, str):
        sources = levels.pre_et_look_levels(sources, bin_length = bin_length)

    bins = compositer.time_bins(timelim, bin_length)

    general_enhancers = enhancers + [rename_vars, fill_attrs, partial(calc_doys, bins = bins), remove_empty_statics, add_constants_new]

    adjusted_timelim = [bins[0], bins[-1]]
    buffered_timelim = [adjusted_timelim[0] - np.timedelta64(3, "D"), 
                        adjusted_timelim[1] + np.timedelta64(3, "D")]

    dss, sources = downloader.collect_sources(folder, sources, latlim, lonlim, buffered_timelim)

    try:
        ds = compositer.main(dss, sources, folder, general_enhancers, bins)
    except Exception as e:
        log.exception("")
        raise e

    t2 = datetime.datetime.now()
    log.sub().info(f"< PRE_ET_LOOK ({str(t2 - t1)})")

    return ds

if __name__ == "__main__":

#     enhancers = "default"
#     diagnostics = None
#     example_source = None
#     bin_length = "DEKAD"
    enhancers = [lapse_rate_to_all]

#     import pywapor

#     folder = r"/Users/hmcoerver/pywapor_notebooks_b"
#     latlim = [28.9, 29.7]
#     lonlim = [30.2, 31.2]
#     timelim = ["2021-07-01", "2021-07-11"]
#     composite_length = "DEKAD"

#     et_look_version = "v2"
#     export_vars = "default"

#     level = "level_1"

#     et_look_sources = pywapor.general.levels.pre_et_look_levels(level)

#     et_look_sources = {k: v for k, v in et_look_sources.items() if k in ["ndvi", "z"]}

    # et_look_sources["ndvi"]["products"] = [
    #     {'source': 'MODIS',
    #         'product_name': 'MOD13Q1.061',
    #         'enhancers': 'default'},
    #     {'source': 'MODIS', 
    #         'product_name': 'MYD13Q1.061', 
    #         'enhancers': 'default'},
    #     {'source': 'PROBAV',
    #         'product_name': 'S5_TOC_100_m_C1',
    #         'enhancers': 'default',
    #         'is_example': True}
    # ]

    # et_look_sources["r0"]["products"] = [
    #     {'source': 'MODIS',
    #         'product_name': 'MCD43A3.061',
    #         'enhancers': 'default'},
    #     {'source': 'PROBAV',
    #         'product_name': 'S5_TOC_100_m_C1',
    #         'enhancers': 'default'}
    # ]

    # se_root_sources = pywapor.general.levels.pre_se_root_levels(level)
    # se_root_sources["ndvi"]["products"] = et_look_sources["ndvi"]["products"]

    # from functools import partial
    # et_look_sources["se_root"]["products"] = [
    #     {'source': partial(pywapor.se_root.se_root, sources = se_root_sources),
    #         'product_name': 'v2',
    #         'enhancers': 'default'},]

    # ds = pywapor.pre_et_look.main(folder, latlim, lonlim, timelim, 
    #                                 sources = et_look_sources,
    #                                 bin_length = composite_length)






from osgeo import gdal
import numpy as np
import xarray as xr

def gap_fill(ds, var, out_var = None, max_search_dist = 8):
    """Apply gdal.FillNoData to every time slice of `var` in `ds`.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset containing var.
    var : str
        Variable name.
    out_var : str, optional
        New variable name in case `var` should not be overwritten, by default None.
    max_search_dist : int, optional
        Control the maximum size of gaps to be filled, by default 8.

    Returns
    -------
    xr.Dataset
        Gap filled dataset.
    """
    # Create an empty xr.DataArray to store our results.
    new_data = xr.ones_like(ds[var]) * np.nan

    # Since gdal.FillNodata can only handle one band at a time, we have to loop over each scene.
    for t in ds.time:

        # Select the data at time t.
        da_slice = ds[var].sel(time = t)

        # Convert the data to a np.array with numeric no-data-value
        shape = da_slice.shape
        data = np.copy(da_slice.values)
        mask = np.isnan(data)
        ndv = -9999
        data[mask] = ndv

        # Create an in-memory gdal.Dataset.
        driver = gdal.GetDriverByName("MEM")
        gdal_ds = driver.Create('', shape[1], shape[0], 1, gdal.GDT_Float32)
        band = gdal_ds.GetRasterBand(1)
        band.SetNoDataValue(ndv)
        band.WriteArray(data)

        # Pass the gdal.Dataset to the gap filling algorithm.
        _ = gdal.FillNodata(targetBand = band, maskBand = None, 
                            maxSearchDist = max_search_dist, smoothingIterations = 0)

        # Read the results and replace the no-data-values again.
        array = band.ReadAsArray()
        array[array == ndv] = np.nan

        # Release the gdal.Dataset.
        gdal_ds = gdal_ds.FlushCache()

        # Put the results back into the xr.DataArray.
        new_data = xr.where(new_data.time == t, array, new_data)

    if not isinstance(out_var, type(None)):
        ds[out_var] = new_data
    else: 
        ds[var] = new_data

    return ds


__all__ = [
    'main', 
    'general', 
    'et_look_v2_v3', 
    'pre_et_look', 
    'et_look', 
    'collect', 
    'post_et_look', 
    'pre_se_root', 
    'se_root', 
    'enhancers'
    ]
__version__ = '3.6.0'
from . import main, general, et_look_v2_v3, pre_et_look, et_look, collect, post_et_look, pre_se_root, se_root, enhancers
from .main import Project, Configuration


import copy
import glob
import os
import warnings
from datetime import datetime as dt
from functools import partial

import numpy as np
import xarray as xr
from lxml import etree

import pywapor.collect.protocol.copernicus_odata as copernicus_odata
from pywapor.enhancers.gap_fill import gap_fill
from pywapor.general.logger import adjust_logger, log
from pywapor.general.processing_functions import adjust_timelim_dtype, open_ds


def apply_qa(ds, var):
    """Mask SENTINEL2 data using a qa variable.

    Parameters
    ----------
    ds : xr.Dataset
        Input data, should contain `"qa"` and `var` as variables.
    var : str
        Variable name in `ds` to be masked.

    Returns
    -------
    xr.Dataset
        Masked dataset.
    """
    # 0 SC_NODATA # 1 SC_SATURATED_DEFECTIVE # 2 SC_DARK_FEATURE_SHADOW
    # 3 SC_CLOUD_SHADOW # 4 SC_VEGETATION # 5 SC_NOT_VEGETATED
    # 6 SC_WATER # 7 SC_UNCLASSIFIED # 8 SC_CLOUD_MEDIUM_PROBA
    # 9 SC_CLOUD_HIGH_PROBA # 10 SC_THIN_CIRRUS # 11 SC_SNOW_ICE
    if "qa" in ds.data_vars:
        pixel_qa_flags = [0, 1, 2, 3, 7, 8, 9, 10, 11]
        keep = np.invert(ds["qa"].isin(pixel_qa_flags))
        ds[var] = ds[var].where(keep)
    else:
        log.warning(
            f"--> Couldn't apply qa, since `qa` doesn't exist in this dataset ({list(ds.data_vars)})."
        )
    return ds


def mask_invalid(ds, var, valid_range=(1, 65534)):
    """Mask invalid data.

    Parameters
    ----------
    ds : xr.Dataset
        Input data.
    var : str
        Variable to mask.
    valid_range : tuple, optional
        Range of valid values in `var`, by default (1, 65534).

    Returns
    -------
    xr.Dataset
        Masked data.
    """
    # 0 = NODATA, 65535 = SATURATED
    ds[var] = ds[var].where((ds[var] >= valid_range[0]) & (ds[var] <= valid_range[1]))
    return ds


def scale_data(ds, var):
    """Apply a scale and offset factor to `ds`.

    Parameters
    ----------
    ds : xr.Dataset
        Input dataset.
    var : str
        Variable to scale and offset.

    Returns
    -------
    xr.Dataset
        Output data.
    """

    scale = 1.0 / ds.scale_factor  # BOA_QUANTIFICATION_VALUE
    offset = ds.offset_factor  # BOA_ADD_OFFSET
    ds[var] = (ds[var] + offset) * scale
    ds[var] = ds[var].where((ds[var] <= 1.00) & (ds[var] >= 0.00))
    return ds


def calc_normalized_difference(ds, var, bands=["nir", "red"]):
    """Calculate the normalized difference of two bands.

    Parameters
    ----------
    ds : xr.Dataset
        Input data.
    var : str
        Name of the variable in which to store the normalized difference.
    bands : list, optional
        The two bands to use to calculate the norm. difference, by default ["nir", "red"].

    Returns
    -------
    xr.Dataset
        Output data.
    """
    if np.all([x in ds.data_vars for x in bands]):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="All-NaN slice encountered")
            da = (ds[bands[0]] - ds[bands[1]]) / (ds[bands[0]] + ds[bands[1]])
        ds[var] = da.clip(-1, 1)
    else:
        log.warning(
            f"--> Couldn't calculate `{var}`, `{'` and `'.join([x for x in bands if x not in ds.data_vars])}` is missing."
        )
    return ds


def calc_psri(ds, var):
    """Calculate the PSRI as ("red" - "blue)/"red_edge_740".

    Parameters
    ----------
    ds : xr.Dataset
        Input data.
    var : str
        Name of the variable in which to store the normalized difference.

    Returns
    -------
    xr.Dataset
        Output data.
    """
    reqs = ["red", "blue", "red_edge_740"]
    if np.all([x in ds.data_vars for x in reqs]):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="All-NaN slice encountered")
            da = (ds["red"] - ds["blue"]) / ds["red_edge_740"]
        ds[var] = da.clip(-1, 1)
    else:
        log.warning(
            f"--> Couldn't calculate `{var}`, `{'` and `'.join([x for x in reqs if x not in ds.data_vars])}` is missing."
        )
    return ds


def calc_nmdi(ds, var):
    """Calculate the NMDI.

    Parameters
    ----------
    ds : xr.Dataset
        Input data.
    var : str
        Name of the variable in which to store the normalized difference.

    Returns
    -------
    xr.Dataset
        Output data.
    """
    reqs = ["swir1", "swir2", "nir"]
    if np.all([x in ds.data_vars for x in reqs]):
        ds["nominator"] = ds["swir1"] - ds["swir2"]
        ds = calc_normalized_difference(ds, var, bands=["nominator", "nir"])
        ds = ds.drop_vars(["nominator"])
    else:
        log.warning(
            f"--> Couldn't calculate `{var}`, `{'` and `'.join([x for x in reqs if x not in ds.data_vars])}` is missing."
        )
    return ds


def calc_bsi(ds, var):
    """Calculate the BSI.

    Parameters
    ----------
    ds : xr.Dataset
        Input data.
    var : str
        Name of the variable in which to store the normalized difference.

    Returns
    -------
    xr.Dataset
        Output data.
    """
    reqs = ["nir", "swir1", "red", "blue"]
    if np.all([x in ds.data_vars for x in reqs]):
        ds["nominator"] = ds["nir"] + ds["blue"]
        ds["denominator"] = ds["swir1"] + ds["red"]
        ds = calc_normalized_difference(ds, var, bands=["nominator", "denominator"])
        ds = ds.drop_vars(["nominator", "denominator"])
    else:
        log.warning(
            f"--> Couldn't calculate `{var}`, `{'` and `'.join([x for x in reqs if x not in ds.data_vars])}` is missing."
        )
    return ds


def calc_vari_red_egde(ds, var):
    """Calculate the VARI_RED_EDGE.

    Parameters
    ----------
    ds : xr.Dataset
        Input data.
    var : str
        Name of the variable in which to store the normalized difference.

    Returns
    -------
    xr.Dataset
        Output data.
    """
    reqs = ["red_edge_740", "blue", "red"]
    if np.all([x in ds.data_vars for x in reqs]):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="All-NaN slice encountered")
            n1 = ds["red_edge_740"] - 1.7 * ds["red"] + 0.7 * ds["blue"]
            n2 = ds["red_edge_740"] + 2.3 * ds["red"] - 1.3 * ds["blue"]
            da = n1 / n2
        ds[var] = da.clip(-1, 1)
    else:
        log.warning(
            f"--> Couldn't calculate `{var}`, `{'` and `'.join([x for x in reqs if x not in ds.data_vars])}` is missing."
        )
    return ds


def calc_r0(ds, var):
    """Calculate the Albedo.

    Parameters
    ----------
    ds : xr.Dataset
        Input data.
    var : str
        Name of the variable in which to store the normalized difference.

    Returns
    -------
    xr.Dataset
        Output data.
    """

    weights = {
        "blue": 0.171,
        "green": 0.060,
        "red": 0.334,
        "nir": 0.371,
        "offset": 0.018,
    }

    reqs = ["blue", "green", "red", "nir"]
    if np.all([x in ds.data_vars for x in reqs]):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="All-NaN slice encountered")
            ds["offset"] = xr.ones_like(ds["blue"])
            weights_da = xr.DataArray(
                data=list(weights.values()), coords={"band": list(weights.keys())}
            )
            ds["r0"] = (
                ds[reqs + ["offset"]]
                .to_array("band")
                .weighted(weights_da)
                .sum("band", skipna=False)
            )
    else:
        log.warning(
            f"--> Couldn't calculate `{var}`, `{'` and `'.join([x for x in reqs if x not in ds.data_vars])}` is missing."
        )
    return ds


def default_vars(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary
    with metadata on which exact layers need to be requested from the server, how they should
    be renamed, and how their dimensions are defined.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Metadata on which exact layers need to be requested from the server.
    """

    variables = {
        "S2MSI2A_R10m": {
            "_B02_10m.jp2": [(), "blue", [mask_invalid, apply_qa, scale_data]],
            "_B03_10m.jp2": [(), "green", [mask_invalid, apply_qa, scale_data]],
            "_B04_10m.jp2": [(), "red", [mask_invalid, apply_qa, scale_data]],
            "_B08_10m.jp2": [(), "nir", [mask_invalid, apply_qa, scale_data]],
            "_SCL_20m.jp2": [(), "qa", []],
        },
        "S2MSI2A_R20m": {
            "_B01_20m.jp2": [
                (),
                "coastal_aerosol",
                [mask_invalid, apply_qa, scale_data],
            ],
            "_B02_20m.jp2": [(), "blue", [mask_invalid, apply_qa, scale_data]],
            "_B03_20m.jp2": [(), "green", [mask_invalid, apply_qa, scale_data]],
            "_B04_20m.jp2": [(), "red", [mask_invalid, apply_qa, scale_data]],
            "_B05_20m.jp2": [(), "red_edge_703", [mask_invalid, apply_qa, scale_data]],
            "_B06_20m.jp2": [(), "red_edge_740", [mask_invalid, apply_qa, scale_data]],
            "_B07_20m.jp2": [(), "red_edge_782", [mask_invalid, apply_qa, scale_data]],
            "_B8A_20m.jp2": [(), "nir", [mask_invalid, apply_qa, scale_data]],
            "_B11_20m.jp2": [(), "swir1", [mask_invalid, apply_qa, scale_data]],
            "_B12_20m.jp2": [(), "swir2", [mask_invalid, apply_qa, scale_data]],
            "_SCL_20m.jp2": [(), "qa", []],
        },
        "S2MSI2A_R60m": {
            "_B01_60m.jp2": [
                (),
                "coastal_aerosol",
                [mask_invalid, apply_qa, scale_data],
            ],
            "_B02_60m.jp2": [(), "blue", [mask_invalid, apply_qa, scale_data]],
            "_B03_60m.jp2": [(), "green", [mask_invalid, apply_qa, scale_data]],
            "_B04_60m.jp2": [(), "red", [mask_invalid, apply_qa, scale_data]],
            "_B05_60m.jp2": [(), "red_edge_703", [mask_invalid, apply_qa, scale_data]],
            "_B06_60m.jp2": [(), "red_edge_740", [mask_invalid, apply_qa, scale_data]],
            "_B07_60m.jp2": [(), "red_edge_782", [mask_invalid, apply_qa, scale_data]],
            "_B8A_60m.jp2": [(), "nir", [mask_invalid, apply_qa, scale_data]],
            "_B09_60m.jp2": [(), "narrow_nir", [mask_invalid, apply_qa, scale_data]],
            "_B11_60m.jp2": [(), "swir1", [mask_invalid, apply_qa, scale_data]],
            "_B12_60m.jp2": [(), "swir2", [mask_invalid, apply_qa, scale_data]],
            "_SCL_60m.jp2": [(), "qa", []],
        },
    }

    req_dl_vars = {
        "S2MSI2A_R10m": {
            "blue": ["_B02_10m.jp2", "_SCL_20m.jp2"],
            "green": ["_B03_10m.jp2", "_SCL_20m.jp2"],
            "red": ["_B04_10m.jp2", "_SCL_20m.jp2"],
            "nir": ["_B08_10m.jp2", "_SCL_20m.jp2"],
            "qa": ["_SCL_20m.jp2"],
            "ndvi": ["_B04_10m.jp2", "_B08_10m.jp2", "_SCL_20m.jp2"],
            "r0": [
                "_B02_10m.jp2",
                "_B03_10m.jp2",
                "_B04_10m.jp2",
                "_B08_10m.jp2",
                "_SCL_20m.jp2",
            ],
        },
        "S2MSI2A_R20m": {
            "coastal_aerosol": ["_B01_20m.jp2", "_SCL_20m.jp2"],
            "blue": ["_B02_20m.jp2", "_SCL_20m.jp2"],
            "green": ["_B03_20m.jp2", "_SCL_20m.jp2"],
            "red": ["_B04_20m.jp2", "_SCL_20m.jp2"],
            "red_edge_703": ["_B05_20m.jp2", "_SCL_20m.jp2"],
            "red_edge_740": ["_B06_20m.jp2", "_SCL_20m.jp2"],
            "red_edge_782": ["_B07_20m.jp2", "_SCL_20m.jp2"],
            "nir": ["_B8A_20m.jp2", "_SCL_20m.jp2"],
            "swir1": ["_B11_20m.jp2", "_SCL_20m.jp2"],
            "swir2": ["_B12_20m.jp2", "_SCL_20m.jp2"],
            "qa": ["_SCL_20m.jp2"],
            "ndvi": ["_B04_20m.jp2", "_B8A_20m.jp2", "_SCL_20m.jp2"],
            "mndwi": ["_B03_20m.jp2", "_B11_20m.jp2", "_SCL_20m.jp2"],
            "vari_red_edge": [
                "_B06_20m.jp2",
                "_B02_20m.jp2",
                "_B04_20m.jp2",
                "_SCL_20m.jp2",
            ],
            "nmdi": ["_B11_20m.jp2", "_B12_20m.jp2", "_B8A_20m.jp2", "_SCL_20m.jp2"],
            "psri": ["_B02_20m.jp2", "_B04_20m.jp2", "_B06_20m.jp2", "_SCL_20m.jp2"],
            "bsi": [
                "_B8A_20m.jp2",
                "_B11_20m.jp2",
                "_B04_20m.jp2",
                "_B02_20m.jp2",
                "_SCL_20m.jp2",
            ],
            "r0": [
                "_B02_20m.jp2",
                "_B03_20m.jp2",
                "_B04_20m.jp2",
                "_B8A_20m.jp2",
                "_SCL_20m.jp2",
            ],
        },
        "S2MSI2A_R60m": {
            "coastal_aerosol": ["_B01_60m.jp2", "_SCL_60m.jp2"],
            "blue": ["_B02_60m.jp2", "_SCL_60m.jp2"],
            "green": ["_B03_60m.jp2", "_SCL_60m.jp2"],
            "red": ["_B04_60m.jp2", "_SCL_60m.jp2"],
            "red_edge_703": ["_B05_60m.jp2", "_SCL_60m.jp2"],
            "red_edge_740": ["_B06_60m.jp2", "_SCL_60m.jp2"],
            "red_edge_782": ["_B07_60m.jp2", "_SCL_60m.jp2"],
            "nir": ["_B8A_60m.jp2", "_SCL_60m.jp2"],
            "narrow_nir": ["_B09_60m.jp2", "_SCL_60m.jp2"],
            "swir1": ["_B11_60m.jp2", "_SCL_60m.jp2"],
            "swir2": ["_B12_60m.jp2", "_SCL_60m.jp2"],
            "qa": ["_SCL_60m.jp2"],
            "ndvi": ["_B04_60m.jp2", "_B8A_60m.jp2", "_SCL_60m.jp2"],
            "mndwi": ["_B03_60m.jp2", "_B11_60m.jp2", "_SCL_60m.jp2"],
            "vari_red_edge": [
                "_B06_60m.jp2",
                "_B02_60m.jp2",
                "_B04_60m.jp2",
                "_SCL_60m.jp2",
            ],
            "nmdi": ["_B11_60m.jp2", "_B12_60m.jp2", "_B8A_60m.jp2", "_SCL_60m.jp2"],
            "psri": ["_B02_60m.jp2", "_B04_60m.jp2", "_B06_60m.jp2", "_SCL_60m.jp2"],
            "bsi": [
                "_B8A_60m.jp2",
                "_B11_60m.jp2",
                "_B04_60m.jp2",
                "_B02_60m.jp2",
                "_SCL_60m.jp2",
            ],
            "r0": [
                "_B02_60m.jp2",
                "_B03_60m.jp2",
                "_B04_60m.jp2",
                "_B8A_60m.jp2",
                "_SCL_60m.jp2",
            ],
        },
    }

    out = {
        val: variables[product_name][val]
        for sublist in map(req_dl_vars[product_name].get, req_vars)
        for val in sublist
    }

    return out


def default_post_processors(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary with a
    list of functions per variable that should be applied after having collected the data
    from a server.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Functions per variable that should be applied to the variable.
    """

    post_processors = {
        "S2MSI2A_R10m": {
            "blue": [gap_fill],
            "green": [gap_fill],
            "red": [gap_fill],
            "nir": [gap_fill],
            "qa": [],
            "ndvi": [calc_normalized_difference, gap_fill],
            "r0": [calc_r0, gap_fill],
        },
        "S2MSI2A_R20m": {
            "coastal_aerosol": [gap_fill],
            "blue": [gap_fill],
            "green": [gap_fill],
            "red": [gap_fill],
            "red_edge_703": [gap_fill],
            "red_edge_740": [gap_fill],
            "red_edge_782": [gap_fill],
            "nir": [gap_fill],
            "qa": [],
            "swir1": [gap_fill],
            "swir2": [gap_fill],
            "psri": [calc_psri],
            "ndvi": [calc_normalized_difference, gap_fill],
            "nmdi": [calc_nmdi, gap_fill],
            "vari_red_edge": [calc_vari_red_egde, gap_fill],
            "bsi": [calc_bsi, gap_fill],
            "mndwi": [
                partial(calc_normalized_difference, bands=["swir1", "green"]),
                gap_fill,
            ],
            "r0": [calc_r0, gap_fill],
        },
        "S2MSI2A_R60m": {
            "coastal_aerosol": [gap_fill],
            "blue": [gap_fill],
            "green": [gap_fill],
            "red": [gap_fill],
            "red_edge_703": [gap_fill],
            "red_edge_740": [gap_fill],
            "red_edge_782": [gap_fill],
            "nir": [gap_fill],
            "narrow_nir": [gap_fill],
            "qa": [],
            "swir1": [gap_fill],
            "swir2": [gap_fill],
            "psri": [calc_psri, gap_fill],
            "ndvi": [calc_normalized_difference, gap_fill],
            "nmdi": [calc_nmdi, gap_fill],
            "vari_red_edge": [calc_vari_red_egde, gap_fill],
            "bsi": [calc_bsi, gap_fill],
            "mndwi": [
                partial(calc_normalized_difference, bands=["swir1", "green"]),
                gap_fill,
            ],
            "r0": [calc_r0, gap_fill],
        },
    }

    out = {k: v for k, v in post_processors[product_name].items() if k in req_vars}

    return out


def time_func(fn):
    """Return a np.datetime64 given a filename.

    Parameters
    ----------
    fn : str
        Filename.

    Returns
    -------
    np.datetime64
        Date as described in the filename.
    """
    dtime = np.datetime64(dt.strptime(fn.split("_")[2], "%Y%m%dT%H%M%S"), "ns")
    return dtime


def s2_processor(scene_folder, variables, **kwargs):
    has_10m = np.any(["_10m" in x for x in variables.keys()])

    dss = list()
    dss__ = list()
    lowres_vars_idxs = list()
    for i, (k, v) in enumerate(variables.items()):
        ds__ = open_ds(
            glob.glob(os.path.join(scene_folder, "**", "*" + k), recursive=True)[0],
            decode_coords=None,
        )
        ds_ = ds__.isel(band=0).rename({"band_data": v[1]})
        dss__.append(ds__)
        dss.append(ds_)
        if has_10m and "_20m" in k:
            lowres_vars_idxs.append(i)

    if len(lowres_vars_idxs) < len(dss) and has_10m:
        highres_ds = dss[[i for i in range(len(dss)) if i not in lowres_vars_idxs][0]]
        for idx in lowres_vars_idxs:
            dss[idx] = dss[idx].interp_like(
                highres_ds, method="nearest", kwargs={"fill_value": "extrapolate"}
            )

    ds = xr.merge(dss).drop_vars("band")

    meta_fps = glob.glob(
        os.path.join(scene_folder, "**", "MTD_MSIL2A.xml"), recursive=True
    )

    if len(meta_fps) >= 1:
        tree = etree.parse(meta_fps[0])
        root = tree.getroot()
        baseline = [float(x.text) for x in root.iter("PROCESSING_BASELINE")][0]
        # NOTE https://sentinels.copernicus.eu/en/web/sentinel/-/copernicus-sentinel-2-major-products-upgrade-upcoming
        # NOTE https://stackoverflow.com/questions/72566760/how-to-correct-sentinel-2-baseline-v0400-offset
        if baseline >= 4.0:
            offsets = [float(x.text) for x in root.iter("BOA_ADD_OFFSET")]
            scales = [float(x.text) for x in root.iter("BOA_QUANTIFICATION_VALUE")]
            scale = np.median(scales)
            offset = np.median(offsets)
            ds.attrs = {"scale_factor": scale, "offset_factor": offset}
        else:
            offsets = [0.0]
            offset = offsets[0]
            scales = [float(x.text) for x in root.iter("BOA_QUANTIFICATION_VALUE")]
            scale = np.median(scales)
            ds.attrs = {"scale_factor": scale, "offset_factor": offset}
        if len(np.unique(offsets)) != 1:
            log.warning(
                f"--> Multiple offsets found for `{scene_folder}`, using `{offset}`, check `{meta_fps[0]}` for more info."
            )
        if len(scales) != 1:
            log.warning(
                f"--> Multiple scales found for `{scene_folder}`, using `{scale}`, check `{meta_fps[0]}` for more info."
            )
    else:
        t = time_func(os.path.split(scene_folder)[-1])
        if t < np.datetime64("2022-01-25"):
            ds.attrs = {"scale_factor": 10000.0, "offset_factor": 0.0}
        else:
            ds.attrs = {"scale_factor": 10000.0, "offset_factor": -1000.0}
        log.warning(
            f"--> No scale/offset found for `{scene_folder}`, using `{ds.attrs}`."
        )

    return ds, dss__


def most_recent(product_name, latlim, lonlim):
    return copernicus_odata.most_recent(product_name, latlim, lonlim, "SENTINEL-2")


def download(
    folder,
    latlim,
    lonlim,
    timelim,
    product_name,
    req_vars,
    variables=None,
    post_processors=None,
    extra_search_kwargs={"cloudcoverpercentage": (0, 30)},
):
    """Download SENTINEL2 data and store it in a single netCDF file.

    Parameters
    ----------
    folder : str
        Path to folder in which to store results.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    product_name : str
        Name of the product to download.
    req_vars : list
        Which variables to download for the selected product.
    variables : dict, optional
        Metadata on which exact layers need to be requested from the server, by default None.
    post_processors : dict, optional
        Functions per variable that should be applied to the variable, by default None.
    extra_search_kwargs : dict
        Extra search kwargs passed to SentinelAPI, by default {"cloudcoverpercentage": (0, 30)}.

    Returns
    -------
    xr.Dataset
        Downloaded data.
    """

    product_folder = os.path.join(folder, "SENTINEL2")

    fn = os.path.join(product_folder, f"{product_name}.nc")
    req_vars_orig = copy.deepcopy(req_vars)
    if os.path.isfile(fn):
        existing_ds = open_ds(fn)
        req_vars_new = list(set(req_vars).difference(set(existing_ds.data_vars)))
        if len(req_vars_new) > 0:
            req_vars = req_vars_new
            existing_ds = existing_ds.close()
        else:
            return existing_ds[req_vars_orig]

    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {
            k: {True: default_processors[k], False: v}[v == "default"]
            for k, v in post_processors.items()
            if k in req_vars
        }

    timelim = adjust_timelim_dtype(timelim)
    bb = [lonlim[0], latlim[0], lonlim[1], latlim[1]]

    def node_filter(node_info):
        fn = os.path.split(node_info)[-1]
        to_dl = list(variables.keys()) + ["MTD_MSIL2A.xml"]
        return np.any([x in fn for x in to_dl])

    final_fn = os.path.split(fn)[-1]
    processor = s2_processor
    scenes = copernicus_odata.download(
        product_folder,
        latlim,
        lonlim,
        timelim,
        "SENTINEL2",
        product_name.split("_")[0],
        node_filter=node_filter,
    )
    ds = copernicus_odata.process_sentinel(
        scenes, variables, time_func, final_fn, post_processors, processor, bb=bb
    )

    return ds[req_vars_orig]


if __name__ == "__main__":
    os.environ["PYWAPOR_REMOVE_TEMP_FILES"] = "NO"

    folder = r"/Users/hmcoerver/Local/s2_test"
    adjust_logger(True, folder, "INFO")
    timelim = ["2022-03-29", "2022-03-31"]
    latlim = [28.9, 29.1]
    lonlim = [30.2, 30.4]

    product_name = "S2MSI2A_R10m"
    # req_vars = ["mndwi", "psri", "vari_red_edge", "bsi", "nmdi", "green", "nir"]
    req_vars = ["ndvi", "r0"]
    post_processors = None
    variables = None

    # ds = download(folder, latlim, lonlim, timelim, product_name,
    #             req_vars, variables = variables, post_processors = post_processors)

    # variables = {'_B02_10m.jp2': [(), 'blue', []], '_SCL_20m.jp2': [(), 'qa', []]}

    # scene_folder = r"/Users/hmcoerver/Local/s2_test/SENTINEL2/S2A_MSIL2A_20220329T082601_N0400_R021_T36RTS_20220329T125458.SAFE"



import os
import requests
import tqdm
from pywapor.general.processing_functions import save_ds, open_ds, process_ds, remove_ds
from osgeo import gdal
import urllib
from pywapor.enhancers.apply_enhancers import apply_enhancers
from pywapor.general.logger import log

def download(fp, product_name, coords, variables, post_processors, url_func, 
                gdal_config_options = {}, waitbar = True, ndv = -9999):
    """Download data from a Cloud Optimized Geotiff hosted on a remote server.

    Parameters
    ----------
    fp : str
        Path to file in which to download.
    product_name : str
        Name of product.
    coords : dict
        Coordinate names and boundaries.
    variables : dict
        Keys are variable names, values are additional settings.
    post_processors : dict
        Processors to apply to specific variables.
    url_func : function
        Function that takes `product_name` as input and return a url.
    gdal_config_options : dict, optional
        Additional options passed to `gdal.SetConfigOption`, by default {}.
    waitbar : bool, optional
        Show a download progress bar or not, by default True.
    ndv : int, optional
        No data value to use in new file, by default -9999.

    Returns
    -------
    xr.Dataset
        Dataset with the downloaded data.
    """

    folder, fn = os.path.split(fp)

    # Create folder.
    if not os.path.isdir(folder):
        os.makedirs(folder)

    try:
        for k, v in gdal_config_options.items():
            gdal.SetConfigOption(k, v)

        bands = [int(k.replace("Band", "")) for k in variables.keys() if "Band" in k]

        # Define bounding-box.
        bb = [coords["x"][1][0], coords["y"][1][1], 
                coords["x"][1][1], coords["y"][1][0]]

        options_dict = {
                "projWin": bb,
                "format": "netCDF",
                "creationOptions": ["COMPRESS=DEFLATE", "FORMAT=NC4"],
                "noData": ndv,
                "bandList": bands,
        }

        if waitbar:
            # Create waitbar.
            waitbar = tqdm.tqdm(position = 0, total = 100, bar_format='{l_bar}{bar}|', delay = 30)

            # Define callback function for waitbar progress.
            def _callback_func(info, *args):
                waitbar.update(info * 100 - waitbar.n)

            # Add callback to gdal.Translate options.
            options_dict.update({"callback": _callback_func})

        # Set gdal.Translate options.
        options = gdal.TranslateOptions(**options_dict)

        # Check if url is local or online path.
        url = url_func(product_name)
        is_not_local = urllib.parse.urlparse(url).scheme in ('http', 'https',)
        if is_not_local and not "vsicurl" in url:
            url = f"/vsicurl/{url}"

        # Check if path is free.
        temp_path = fp.replace(".nc", "_temp.nc")
        while os.path.isfile(temp_path):
            temp_path = temp_path.replace(".nc", "_.nc")

        # Run gdal.Translate.
        ds_ = gdal.Translate(temp_path, url, options = options)
        ds_.FlushCache()
        ds_ = None

    except Exception as e:
        raise e
    finally:
        for k in gdal_config_options.keys():
            gdal.SetConfigOption(k, None)

    # Process the new netCDF.
    ds_ = open_ds(temp_path)

    ds = ds_.rename_vars({k: f"Band{v}" for k,v in zip(list(ds_.data_vars), bands)})

    ds = process_ds(ds, coords, variables)

    # Apply product specific functions.
    ds = apply_enhancers(post_processors, ds)

    # Save final output.
    out = save_ds(ds, temp_path.replace("_temp", ""), encoding = "initiate", label = f"Saving {fn}.")

    # Remove the temporary file.
    remove_ds(ds_)

    return out

def cog_dl(urls, out_fn, overview = "NONE", warp_kwargs = {}, vrt_options = {"separate": True}):

    out_ext = os.path.splitext(out_fn)[-1]
    valid_ext = {".nc": "netCDF", ".tif": "GTiff"}
    valid_cos = {".nc": ["COMPRESS=DEFLATE", "FORMAT=NC4C"], ".tif": ["COMPRESS=LZW"]}
    vrt_fn = out_fn.replace(out_ext, ".vrt")

    log.info("--> Building `.vrt` file.")
    ## Build VRT with all the required data.
    vrt_options_ = gdal.BuildVRTOptions(
        **vrt_options
    )
    vrt = gdal.BuildVRT(vrt_fn, urls, options = vrt_options_)
    vrt.FlushCache()

    ## Download the data.
    warp_options = gdal.WarpOptions(
        format = valid_ext[out_ext],
        cropToCutline = True,
        overviewLevel = overview,
        multithread = True,
        creationOptions = valid_cos[out_ext],
        **warp_kwargs,
    )
    log.info(f"--> Downloading {len(urls)} bands.")
    warp = gdal.Warp(out_fn, vrt_fn, options = warp_options)
    warp.SetMetadataItem('pyWaPOR_bb', os.environ.get("pyWaPOR_bb", "unknown"))
    warp.SetMetadataItem('pyWaPOR_period', os.environ.get("pyWaPOR_period", "unknown"))
    warp.FlushCache() # NOTE do not remove this.

    if os.path.isfile(vrt_fn):
        remove_ds(vrt_fn)

    return out_fn

def ping_urls(urls):

    faulty_urls = list()
    for url in urls:
        resp = requests.get(
            url.replace("/vsicurl/", ""), 
            stream=True
            )
        if not resp.ok:
            faulty_urls.append(url)

    if len(faulty_urls) > 0:
        print_urls = "\n".join(faulty_urls)
        log.warning(f"The following URLs are invalid: {print_urls}")

    return faulty_urls


import copy
import os
import warnings
from functools import partial

import numpy as np
import pandas as pd
import xarray as xr

from pywapor.collect.protocol import opendap
from pywapor.collect.protocol.projections import get_crss
from pywapor.enhancers.pressure import pa_to_kpa
from pywapor.enhancers.temperature import kelvin_to_celsius
from pywapor.enhancers.wind import adjust_wind_height
from pywapor.general.processing_functions import open_ds


def default_vars(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary
    with metadata on which exact layers need to be requested from the server, how they should
    be renamed, and how their dimensions are defined.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Metadata on which exact layers need to be requested from the server.
    """

    variables = {
        "inst3_2d_asm_Nx": {
            "t2m": [("time", "lat", "lon"), "t_air"],
            "u2m": [("time", "lat", "lon"), "u2m"],
            "v2m": [("time", "lat", "lon"), "v2m"],
            "qv2m": [("time", "lat", "lon"), "qv"],
            "tqv": [("time", "lat", "lon"), "wv"],
            "ps": [("time", "lat", "lon"), "p_air"],
            "slp": [("time", "lat", "lon"), "p_air_0"],
        },
        "tavg1_2d_slv_Nx": {
            "t2m": [("time", "lat", "lon"), "t_air"],
            "qv2m": [("time", "lat", "lon"), "qv"],
            "u10m": [("time", "lat", "lon"), "u10m"],
            "v10m": [("time", "lat", "lon"), "v10m"],
            "ps": [("time", "lat", "lon"), "p_air"],
            "slp": [("time", "lat", "lon"), "p_air_0"],
            "tqv": [("time", "lat", "lon"), "wv"],
            "to3": [("time", "lat", "lon"), "to3"],
        },
        "tavg1_2d_rad_Nx": {
            "swgdn": [
                ("time", "lat", "lon"),
                "ra_flat",
            ],  # W/m2, surface incoming shortwave flux
        },
        "tavg1_2d_lnd_Nx": {
            "prectot": [("time", "lat", "lon"), "p"],
        },
        "tavg3_2d_aer_Nx": {
            "totangstr": [("time", "lat", "lon"), "totangstr"],
            "totexttau": [("time", "lat", "lon"), "totexttau"],
        },
    }

    req_dl_vars = {
        "inst3_2d_asm_Nx": {
            "t_air": ["t2m"],
            "t_air_max": ["t2m"],
            "t_air_min": ["t2m"],
            "u2m": ["u2m"],
            "v2m": ["v2m"],
            "qv": ["qv2m"],
            "wv": ["tqv"],
            "p_air": ["ps"],
            "p_air_0": ["slp"],
        },
        "tavg1_2d_slv_Nx": {
            "t_air": ["t2m"],
            "t_air_max": ["t2m"],
            "t_air_min": ["t2m"],
            "u10m": ["u10m"],
            "v10m": ["v10m"],
            "u2m": ["u10m"],
            "v2m": ["v10m"],
            "qv": ["qv2m"],
            "wv": ["tqv"],
            "p_air": ["ps"],
            "p_air_0": ["slp"],
            "to3": ["to3"],  # total_column_ozone
        },
        "tavg1_2d_rad_Nx": {
            "ra_flat": ["swgdn"],  # surface_incoming_shortwave_flux
        },
        "tavg1_2d_lnd_Nx": {
            "p": ["prectot"],
        },
        "tavg3_2d_aer_Nx": {
            "totangstr": ["totangstr"],  # total aerosol angstrom parameter [470-870 nm]
            "totexttau": ["totexttau"],  # total aerosol extinction aot [550 nm]
        },
    }

    out = {
        val: variables[product_name][val]
        for sublist in map(req_dl_vars[product_name].get, req_vars)
        for val in sublist
    }

    return out


def default_post_processors(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary with a
    list of functions per variable that should be applied after having collected the data
    from a server.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Functions per variable that should be applied to the variable.
    """

    post_processors = {
        "inst3_2d_asm_Nx": {
            "t_air": [kelvin_to_celsius],
            "t_air_max": [
                partial(kelvin_to_celsius, in_var="t_air", out_var="t_air_max")
            ],
            "t_air_min": [
                partial(kelvin_to_celsius, in_var="t_air", out_var="t_air_min")
            ],
            "u2m": [],
            "v2m": [],
            "qv": [],
            "wv": [],
            "p_air": [pa_to_kpa],
            "p_air_0": [pa_to_kpa],
        },
        "tavg1_2d_slv_Nx": {
            "t_air": [kelvin_to_celsius],
            "t_air_max": [
                partial(kelvin_to_celsius, in_var="t_air", out_var="t_air_max")
            ],
            "t_air_min": [
                partial(kelvin_to_celsius, in_var="t_air", out_var="t_air_min")
            ],
            "v10m": [],
            "u10m": [],
            "u2m": [adjust_wind_height],
            "v2m": [adjust_wind_height],
            "qv": [],
            "wv": [],
            "p_air": [pa_to_kpa],
            "p_air_0": [pa_to_kpa],
            "to3": [],
        },
        "tavg1_2d_rad_Nx": {
            "ra_flat": [],
        },
        "tavg1_2d_lnd_Nx": {
            "p": [],
        },
        "tavg3_2d_aer_Nx": {
            "totangstr": [],
            "totexttau": [],
        },
    }

    out = {k: v for k, v in post_processors[product_name].items() if k in req_vars}

    return out


def most_recent(product_name, *args):
    url = f"https://opendap.nccs.nasa.gov/dods/GEOS-5/fp/0.25_deg/assim/{product_name}"

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=xr.SerializationWarning)
        x = xr.open_dataset(url)

    return pd.to_datetime(x["time"][-1].values).to_pydatetime()


def download(
    folder,
    latlim,
    lonlim,
    timelim,
    product_name,
    req_vars,
    variables=None,
    post_processors=None,
):
    """Download GEOS5 data and store it in a single netCDF file. Product
    docs are here https://gmao.gsfc.nasa.gov/pubs/docs/Lucchesi1202.pdf

    Parameters
    ----------
    folder : str
        Path to folder in which to store results.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    product_name : str
        Name of the product to download.
    req_vars : list
        Which variables to download for the selected product.
    variables : dict, optional
        Metadata on which exact layers need to be requested from the server, by default None.
    post_processors : dict, optional
        Functions per variable that should be applied to the variable, by default None.

    Returns
    -------
    xr.Dataset
        Downloaded data.
    """
    folder = os.path.join(folder, "GEOS5")

    if not os.path.exists(folder):
        os.makedirs(folder)

    fn = os.path.join(folder, f"{product_name}.nc")
    req_vars_orig = copy.deepcopy(req_vars)
    if os.path.isfile(fn):
        existing_ds = open_ds(fn)
        req_vars_new = list(set(req_vars).difference(set(existing_ds.data_vars)))
        if len(req_vars_new) > 0:
            req_vars = req_vars_new
            existing_ds = existing_ds.close()
        else:
            return existing_ds[req_vars_orig]

    spatial_buffer = True
    if spatial_buffer:
        latlim = [latlim[0] - 0.25, latlim[1] + 0.25]
        lonlim = [lonlim[0] - 0.3125, lonlim[1] + 0.3125]

    coords = {"x": ["lon", lonlim], "y": ["lat", latlim], "t": ["time", timelim]}

    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {
            k: {True: default_processors[k], False: v}[v == "default"]
            for k, v in post_processors.items()
            if k in req_vars
        }

    data_source_crs = get_crss("WGS84")

    url = f"https://opendap.nccs.nasa.gov/dods/GEOS-5/fp/0.25_deg/assim/{product_name}"

    timedelta = np.timedelta64(90, "m")

    ds = opendap.download_xarray(
        url,
        fn,
        coords,
        variables,
        post_processors,
        data_source_crs=data_source_crs,
        timedelta=timedelta,
    )

    return ds[req_vars_orig]


if __name__ == "__main__":
    import datetime

    args = {
        "folder": "/Users/hmcoerver/Local/test_dl_GEOS5_0",
        "latlim": [29.4, 29.7],
        "lonlim": [30.7, 31.0],
        "timelim": [datetime.date(2022, 3, 1), datetime.date(2022, 3, 3)],
        "product_name": "inst3_2d_asm_Nx",
        "req_vars": [
            "t_air",
            "t_air_max",
            "t_air_min",
            "u2m",
            "v2m",
            "qv",
            "wv",
            "p_air",
            "p_air_0",
        ],
    }

    variables = None
    post_processors = None

    for var, value in args.items():
        if isinstance(value, str):
            exec(f"{var} = '{value}'")
        else:
            exec(f"{var} = {value}")

    # adjust_logger(True, folder, "INFO")



import copy
import datetime
import fnmatch
import os
from functools import partial
from urllib.error import HTTPError

import numpy as np
import pandas as pd

import pywapor.collect.accounts as accounts
import pywapor.collect.protocol.opendap as opendap
from pywapor.collect.protocol.crawler import find_paths
from pywapor.collect.protocol.projections import get_crss
from pywapor.enhancers.pressure import pa_to_kpa
from pywapor.enhancers.temperature import kelvin_to_celsius
from pywapor.general.processing_functions import open_ds


def default_vars(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary
    with metadata on which exact layers need to be requested from the server, how they should
    be renamed, and how their dimensions are defined.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Metadata on which exact layers need to be requested from the server.
    """
    variables = {
        "M2I1NXASM.5.12.4": {
            "T2M": [("time", "lat", "lon"), "t_air"],
            "U2M": [("time", "lat", "lon"), "u2m"],
            "V2M": [("time", "lat", "lon"), "v2m"],
            "QV2M": [("time", "lat", "lon"), "qv"],
            "TQV": [("time", "lat", "lon"), "wv"],
            "PS": [("time", "lat", "lon"), "p_air"],
            "SLP": [("time", "lat", "lon"), "p_air_0"],
        },
        "M2T1NXRAD.5.12.4": {
            "SWGDN": [("time", "lat", "lon"), "ra_flat"],
        },
    }

    req_dl_vars = {
        "M2I1NXASM.5.12.4": {
            "t_air": ["T2M"],
            "t_air_max": ["T2M"],
            "t_air_min": ["T2M"],
            "u2m": ["U2M"],
            "v2m": ["V2M"],
            "qv": ["QV2M"],
            "wv": ["TQV"],
            "p_air": ["PS"],
            "p_air_0": ["SLP"],
        },
        "M2T1NXRAD.5.12.4": {
            "ra_flat": ["SWGDN"],
        },
    }

    out = {
        val: variables[product_name][val]
        for sublist in map(req_dl_vars[product_name].get, req_vars)
        for val in sublist
    }

    return out


def default_post_processors(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary with a
    list of functions per variable that should be applied after having collected the data
    from a server.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Functions per variable that should be applied to the variable.
    """

    post_processors = {
        "M2I1NXASM.5.12.4": {
            "t_air": [kelvin_to_celsius],
            "t_air_max": [
                partial(kelvin_to_celsius, in_var="t_air", out_var="t_air_max")
            ],
            "t_air_min": [
                partial(kelvin_to_celsius, in_var="t_air", out_var="t_air_min")
            ],
            "u2m": [],
            "v2m": [],
            "qv": [],
            "wv": [],
            "p_air": [pa_to_kpa],
            "p_air_0": [pa_to_kpa],
        },
        "M2T1NXRAD.5.12.4": {
            "ra_flat": [],
        },
    }

    out = {k: v for k, v in post_processors[product_name].items() if k in req_vars}

    return out


def fn_func(product_name, tile):
    """Returns a client-side filename at which to store data.

    Parameters
    ----------
    product_name : str
        Name of the product to download.
    tile : str
        Name of the server-side tile to download.

    Returns
    -------
    str
        Filename.
    """
    fn = f"{product_name}_{tile.strftime('%Y%m%d')}.nc"
    return fn


def url_func(product_name, tile):
    """Returns a url at which to collect MERRA2 data.

    Parameters
    ----------
    product_name : str
        Name of the product to download.
    tile : str
        Name of the server-side tile to download.

    Returns
    -------
    str
        The url.
    """

    def _filter(tag):
        tag_value = tag["href"]
        if tag_value[-5:] == ".html":
            tag_value = tag_value[:-5]
        return tag_value

    # Find the existing tiles for the given year and month, this is necessary
    # because the version number (`\d{3}`) in the filenames is irregular.
    regex = r"MERRA2_\d{3}\..*\.\d{8}.nc4.html"
    url = f"https://goldsmr4.gesdisc.eosdis.nasa.gov/opendap/MERRA2/{product_name}/{tile.year}/{tile.month:02d}/contents.html"
    tile_names = find_paths(url, regex, filter=_filter)

    # Find which of the existing tiles matches with the date.
    fn_pattern = f"MERRA2_*.*.{tile.strftime('%Y%m%d')}.nc4"
    fn = fnmatch.filter(tile_names, fn_pattern)[0]

    # Create the final url.
    url = f"https://goldsmr4.gesdisc.eosdis.nasa.gov/opendap/MERRA2/{product_name}/{tile.year}/{tile.month:02d}/{fn}.nc4?"
    return url


def most_recent(product_name, *args):
    today = pd.Timestamp.today().date()
    month_year = today + pd.offsets.MonthBegin()

    def _filter(tag):
        tag_value = tag["href"]
        if tag_value[-5:] == ".html":
            tag_value = tag_value[:-5]
        return tag_value

    regex = r"MERRA2_\d{3}\..*\.\d{8}.nc4.html"

    found = False
    iter = 0
    while not found and iter < 1000:
        url = f"https://goldsmr4.gesdisc.eosdis.nasa.gov/opendap/MERRA2/{product_name}/{month_year.year}/{month_year.month:02d}/contents.html"
        try:
            tile_names = find_paths(url, regex, filter=_filter)
        except HTTPError:
            tile_names = []
        if len(tile_names) == 0:
            month_year -= pd.offsets.MonthBegin()
        else:
            found = True
        iter += 1

    x = datetime.datetime.strptime(sorted(tile_names)[-1].split(".")[-2], "%Y%m%d")
    return x


def download(
    folder,
    latlim,
    lonlim,
    timelim,
    product_name,
    req_vars,
    variables=None,
    post_processors=None,
):
    """Download MERRA2 data and store it in a single netCDF file. See
    https://gmao.gsfc.nasa.gov/pubs/docs/Bosilovich785.pdf for docs.

    Parameters
    ----------
    folder : str
        Path to folder in which to store results.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    product_name : str
        Name of the product to download.
    req_vars : list
        Which variables to download for the selected product.
    variables : dict, optional
        Metadata on which exact layers need to be requested from the server, by default None.
    post_processors : dict, optional
        Functions per variable that should be applied to the variable, by default None.

    Returns
    -------
    xr.Dataset
        Downloaded data.
    """

    folder = os.path.join(folder, "MERRA2")

    fn = os.path.join(folder, f"{product_name}.nc")
    req_vars_orig = copy.deepcopy(req_vars)
    if os.path.isfile(fn):
        existing_ds = open_ds(fn)
        req_vars_new = list(set(req_vars).difference(set(existing_ds.data_vars)))
        if len(req_vars_new) > 0:
            req_vars = req_vars_new
            existing_ds = existing_ds.close()
        else:
            return existing_ds[req_vars_orig]

    spatial_buffer = True
    if spatial_buffer:
        latlim = [latlim[0] - 0.5, latlim[1] + 0.5]
        lonlim = [lonlim[0] - 0.625, lonlim[1] + 0.625]

    tiles = pd.date_range(timelim[0], timelim[1], freq="D")
    coords = {"x": ["lon", lonlim], "y": ["lat", latlim], "t": ["time", timelim]}
    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {
            k: {True: default_processors[k], False: v}[v == "default"]
            for k, v in post_processors.items()
            if k in req_vars
        }

    timedelta = np.timedelta64(30, "m")
    data_source_crs = get_crss("WGS84")
    parallel = False
    spatial_tiles = False
    un_pw = accounts.get("NASA")
    request_dims = True

    ds = opendap.download(
        fn,
        product_name,
        coords,
        variables,
        post_processors,
        fn_func,
        url_func,
        un_pw=un_pw,
        tiles=tiles,
        data_source_crs=data_source_crs,
        parallel=parallel,
        spatial_tiles=spatial_tiles,
        request_dims=request_dims,
        timedelta=timedelta,
    )

    return ds[req_vars_orig]


if __name__ == "__main__":
    folder = r"/Users/hmcoerver/Downloads/pywapor_test"
    # latlim = [26.9, 33.7]
    # lonlim = [25.2, 37.2]
    latlim = [28.9, 29.7]
    lonlim = [30.2, 31.2]
    timelim = [datetime.date(2020, 7, 1), datetime.date(2020, 7, 11)]

    variables = None
    post_processors = None

    # # MERRA2.
    # wanted = [
    #             ("M2I1NXASM.5.12.4", ["t_air", "u2m", "v2m", "qv", "wv", "p_air", "p_air_0"]),
    #             # ("M2T1NXRAD.5.12.4", ["ra"]),
    #         ]

    product_name = "M2I1NXASM.5.12.4"
    req_vars = ["t_air", "u2m", "v2m", "qv", "wv", "p_air", "p_air_0"]

    # for product_name, req_vars in wanted:
    # ds = download(folder, latlim, lonlim, timelim, product_name, req_vars)
    # print(ds.rio.crs, ds.rio.grid_mapping)



"""
https://pypi.org/project/python_log_indenter/
https://github.com/dstrohl/Python_log_indenter/tree/master
https://python-log-indenter.readthedocs.io/en/latest/
"""

__author__ = 'Dan Strohl'
__all__ = ['NamedFiloQueue', 'IndentedLoggerAdapter']


# from logging import LoggerAdapter
import logging


class NamedFiloQueue(object):
    """
    This class is a helper class for the IndentHelperClass.  This manages a FILO queue with named slots.
    """

    def __init__(self, empty_answer=None):
        """
        :param empty_answer: What should we return if the queue is empty?
        """
        self._queue_data = []
        self._pos_names = []
        self._empty_answer = empty_answer

    def push(self, data, name=None):
        """
        Add an item to the queue
        :param data: The data to add
        :param name: The name of the slot (optional)
        :return: This returns itself so that it can be chained.
        """
        if name is not None and not isinstance(name, str):
            raise AttributeError('name parameter must be a string')
        self._queue_data.append(data)
        self._pos_names.append(name)
        return self

    def pop(self, name=None):
        """
        Pull the top item from the queue, or the named or selected item.
        :param name: If this is a string, it will pop all items up to that item, if it is an integer, it will pop that
            many items.
        :return: The data for the last pop'ed item.
        :raises: KeyError if the name does not match one in the list.
        """
        if name is None:
            try:
                self._pos_names.pop()
                return self._queue_data.pop()
            except IndexError:
                return self._empty_answer
        else:
            if isinstance(name, str):
                if name in self._pos_names:
                    try:
                        while True:
                            tmp_ret = self._queue_data.pop()
                            if self._pos_names.pop() == name:
                                return tmp_ret
                    except IndexError:
                        return self._empty_answer
                else:
                    msg = 'Error: key {} not found in queue'.format(name)
                    raise KeyError(msg)
            elif isinstance(name, int):
                for i in range(name):
                    try:
                        tmp_ret = self._queue_data.pop()
                        self._pos_names.pop()
                    except IndexError:
                        tmp_ret = self._empty_answer
                return tmp_ret
            else:
                raise AttributeError('POP must pass a string or int')

    def reset(self):
        """
        Will clear the queue
        """
        self._queue_data = []
        self._pos_names = []
        return self

    def dump(self, format_str='Pos: {pos} / Name: {name} / Value: {value}'):
        """
        Returns the current queue in string format, normally used for troubleshooting.
        :param format_str: a string used by str.format() to format the returned content.
        :return:
        """
        tmp_ret = []
        if self._queue_data:

            for i in range(len(self), 0, -1):
                tmp_ret.append(format_str.format(
                    pos=i-1,
                    name=self._pos_names[i-1],
                    value=self._queue_data[i-1],
                ))

        else:
            tmp_ret.append('NO QUEUE DATA')
        return '\n'.join(tmp_ret)

    def __call__(self):
        if self._queue_data:
            return self._queue_data[len(self._queue_data)-1]
        else:
            return self._empty_answer

    def __len__(self):
        return len(self._queue_data)

    def __contains__(self, item):
        return item in self._pos_names


class IndentHelperMixin(object):

    def __init__(self, *args, **kwargs):
        """
        :keyword int spaces: Sets the number of spaces for each indent
        :keyword str indent_char: Sets the character or string used for each indent space.
        """

        spaces = kwargs.pop('spaces', 4)
        indent_char = kwargs.pop('indent_char', ' ')

        self._current_indent = 0
        self._indent_spaces = spaces
        self._indent_char = indent_char
        self._indent_mems = {}
        self._pop_queue = NamedFiloQueue(empty_answer=0)
        self._string_lookup = {}
        self._trouble_flag = False

        super(IndentHelperMixin, self).__init__(*args, **kwargs)

    '''
    ====================================
    indent control methods
    ====================================
    '''

    def indent_set(self, indent=0):
        """
        Sets the indent to this level

        :param int indent: the indent level
        :return: Self
        """
        self._current_indent = indent
        return self

    def add(self, indent=1):
        """
        Adds this to the indent level

        :param int indent: the number to add to the indent level
        :return: Self
        """
        self._current_indent += indent
        return self

    def sub(self, indent=1):
        """
        Subtracts this from the indent level

        :param int indent: the number to subtract from the indent level
        :return: Self
        """
        self._current_indent -= indent
        if self._current_indent < 0:
            self._current_indent = 0
        return self

    def indent_get(self):
        return self._current_indent

    @property
    def indent_level(self):
        """
        Returns the current indent level

        :rtype: int
        """
        return self._current_indent

    '''
    ====================================
    memory control methods
    ====================================
    '''

    def mem_save(self, key, indent=None):
        """
        Saves an indent to a specified memory key

        :param str key: The key / name to save to
        :param int indent: The indent to be saved, if None, this saves the current indent.
        :return: Self
        """
        if indent is not None:
            tmp_indent = indent
        else:
            tmp_indent = self._current_indent

        self._indent_mems[key] = tmp_indent

        return self

    def mem(self, key):
        """
        Sets the current indent to the memory at location "key"

        :param str key: The key / name to pull from
        :return: Self
        """
        self._current_indent = self._indent_mems[key]
        return self

    def mem_clear(self, key):
        """
        Removes (deletes) the memory location at the specified key.

        :param str key: The key / name to delete.
        :return: Self
        """
        if key in self._indent_mems:
            del self._indent_mems[key]

    '''
    ====================================
    push-pop methods
    ====================================
    '''

    def push(self, *args):
        """
        Saves an indent to the top of a FILO queue.

        This takes 0-2 arguments, if arguments are passed, the argument that is a string willl be used as the key,
        the argument that is an int will be used as the indent.  if two strings or two int's are passed, unpredictable
        results will occur.

        :argument str key: The name to save to, if no string is passed, this will add an un-named item to the queue.
        :param int indent: The indent to be saved, if None, this saves the current indent.
        :return: Self
        """

        new_indent = self._current_indent
        name = None

        if len(args) > 2:
            raise AttributeError('Only 2 arguments can be passed to PUSH')
        else:
            for arg in args:
                if isinstance(arg, str):
                    name = arg
                elif isinstance(arg, int):
                    new_indent = arg

        if new_indent is None:
            new_indent = self._current_indent

        self._pop_queue.push(new_indent, name=name)
        return self

    def pop(self, name=None):
        """
        Returns an indent to the top of a FILO queue.

        :param name: This behaives differently depending on what type of data is passed:

            int:
                If an int is passed, this will pop that many items from the top of the queue.
            str:
                If a string is passed, this will pop all items from the top of the queue until it reaches that named
                item.  if that item does not exist, it will raise a KeyError.
        :return: Self
        """
        self._current_indent = self._pop_queue.pop(name=name)
        return self

    '''
    ====================================
    support methods
    ====================================
    '''

    def debug_indent(self):
        tmp_ret = ['Indent Spaces  : ', self._indent_spaces,
                   'Current Indent : ', self._current_indent,
                   'Memories       : ', self._indent_mems,
                   'Push Queue     : \n', self._pop_queue.dump()]

        return '\n'.join(tmp_ret)

    '''
    ====================================
    printing methods
    ====================================
    '''

    def indent(self, indent=None, indent_char=None, mem=None, pop=None, indent_len=None):

        if indent_len is None:
            indent_len = self._indent_spaces
        if pop is not None:
            self.pop(pop)
        if mem is not None:
            self.mem(mem)
        if indent_char is None:
            indent_char = self._indent_char

        if indent is None:
            indent = self._current_indent

        ind = indent_len * indent
        tmp_ret = ''
        for i in range(ind):
            tmp_ret += indent_char

        return tmp_ret

    def indent_str(self):
        return self.indent()

    '''
    ====================================
    magic methods
    ====================================
    '''

    def __call__(self, indent=None):
        if indent is not None:
            if isinstance(indent, int):
                self.add(indent)

            elif isinstance(indent, str):
                if indent in self._indent_mems:
                    j = self[indent]

                elif indent in self._pop_queue:
                    self.pop(indent)

        return self.indent()

    def __iadd__(self, other):
        self._current_indent += other
        return self

    def __isub__(self, other):
        self._current_indent -= other
        return self

    def __len__(self):
        return self._current_indent

    def __getitem__(self, item):
        self.mem(item)
        return self.indent()

    def __setitem__(self, key, value):
        self.mem_save(key, value)

    def __delitem__(self, key):
        self.mem_clear(key)

    def __int__(self):
        return self._current_indent

    '''
    ====================================
    Shortcuts
    ====================================
    '''

    @property
    def __(self):
        return self.indent()

    def i(self, indent=0):
        """
        Shortcut method for :py:meth:`IndentedLoggerAdapter.indent_set`
        """
        return self.indent_set(indent)

    def a(self, indent=1):
        """
        Shortcut method for :py:meth:`IndentedLoggerAdapter.add`
        """
        return self.add(indent)

    def s(self, indent=1):
        """
        Shortcut method for :py:meth:`IndentedLoggerAdapter.sub`
        """
        return self.sub(indent)

    def ms(self, key=None, indent=None):
        """
        Shortcut method for :py:meth:`IndentedLoggerAdapter.mem_save`
        """
        return self.mem_save(key, indent)

    def m(self, key=None):
        """
        Shortcut method for :py:meth:`IndentedLoggerAdapter.mem`
        """
        return self.mem(key)

    def mc(self, key=None):
        """
        Shortcut method for :py:meth:`IndentedLoggerAdapter.mem_clear`
        """
        return self.mem_clear(key)

    def po(self, name=None):
        """
        Shortcut method for :py:meth:`IndentedLoggerAdapter.pop`
        """
        return self.pop(name)

    def pu(self, *args):
        """
        Shortcut method for :py:meth:`IndentedLoggerAdapter.push`
        """
        return self.push(*args)


class IndentHelper(IndentHelperMixin):

    def __init__(self, *args, **kwargs):
        super(IndentHelper, self).__init__(*args, **kwargs)

    def __str__(self):
        return self.indent()

    def __repr__(self):
        tmp_ret = 'SpaceHelper, level {} indent'.format(self._current_indent)
        return tmp_ret


class IndentedLoggerAdapter(IndentHelperMixin, logging.LoggerAdapter):
    """

    """

    def __init__(self, logger, extra=None, auto_add=True, **kwargs):
        """
        Overrides the init and parameters from :py:class:`logging.Logger`, however adds it adds some functions.

        :param bool auto_add: this controls how the indent is handled, if this is True (default) the indent is automatically
            added to the beginning of the message before it is passed to the logging engine.  If False, the indent is
            passed as part of the :py:class:`logging.LogRecord` as the parameter .indent_str.  this allows access to
            the indent by custom formatting commands if needed.

        :param int spaces: Sets the number of spaces for each indent
        :param str indent_char: Sets the character or string used for each indent space.

        The other thing that this adds is the ability to chain some of the key methods so that it can be combined in
        single lines of code.  the methods that have been modified are:

        * :py:meth:`IndentedLoggerAdapter.debug`
        * :py:meth:`IndentedLoggerAdapter.info`
        * :py:meth:`IndentedLoggerAdapter.warning`
        * :py:meth:`IndentedLoggerAdapter.error`
        * :py:meth:`IndentedLoggerAdapter.critical`
        * :py:meth:`IndentedLoggerAdapter.exception`
        * :py:meth:`IndentedLoggerAdapter.log`
        """
        super(IndentedLoggerAdapter, self).__init__(logger, extra or {}, **kwargs)
        self._current_level = None
        self._auto_add = auto_add

    def process(self, msg, kwargs):
        if self._auto_add:
            msg = str(self.__)+msg
        else:
            if 'extra' not in kwargs:
                kwargs['extra'] = {}

            kwargs['extra']['indent_level'] = str(self.indent_level)
            kwargs['extra']['indent_str'] = str(self.__)

        return msg, kwargs

    def debug(self, msg, *args, **kwargs):
        super(IndentedLoggerAdapter, self).debug(msg, *args, **kwargs)
        return self

    def info(self, msg, *args, **kwargs):
        super(IndentedLoggerAdapter, self).info(msg, *args, **kwargs)
        return self

    def warning(self, msg, *args, **kwargs):
        super(IndentedLoggerAdapter, self).warning(msg, *args, **kwargs)
        return self

    def error(self, msg, *args, **kwargs):
        super(IndentedLoggerAdapter, self).error(msg, *args, **kwargs)
        return self

    def exception(self, msg, *args, **kwargs):
        super(IndentedLoggerAdapter, self).exception(msg, *args, **kwargs)
        return self

    def critical(self, msg, *args, **kwargs):
        super(IndentedLoggerAdapter, self).critical(msg, *args, **kwargs)
        return self

    def log(self, level, msg, *args, **kwargs):
        super(IndentedLoggerAdapter, self).log(level, msg, *args, **kwargs)
        return self


__all__ = [
    'gap_fill', 
    'apply_enhancers', 
    'dem', 
    'lulc', 
    'pressure', 
    'temperature', 
    'other', 
    'dms'
    ]
from . import gap_fill, apply_enhancers, dem, lulc, pressure, temperature, other, dms


"""
Code to download data using the Copernicus OData API.
https://documentation.dataspace.copernicus.eu/APIs/OData.html
"""

import datetime
import os
import re
import shutil
import time
import warnings

import numpy as np
import pandas as pd
import rasterio.crs
import requests
import tqdm
import xarray as xr
from cachetools import TTLCache, cached
from joblib import Memory

from pywapor.collect.accounts import get
from pywapor.collect.protocol.crawler import download_urls
from pywapor.enhancers.apply_enhancers import apply_enhancer, apply_enhancers
from pywapor.general.logger import adjust_logger, log
from pywapor.general.processing_functions import (
    make_example_ds,
    open_ds,
    remove_ds,
    save_ds,
    unpack,
)


@cached(cache=TTLCache(maxsize=2048, ttl=100))
def get_access_token():
    """
    Create account at https://dataspace.copernicus.eu/
    """

    username, password = get("COPERNICUS_DATA_SPACE")

    data = {
        "client_id": "cdse-public",
        "username": username,
        "password": password,
        "grant_type": "password",
    }

    try:
        log.info("--> Requesting access token.")
        r = requests.post(
            "https://identity.dataspace.copernicus.eu/auth/realms/CDSE/protocol/openid-connect/token",
            data=data,
        )
        r.raise_for_status()
    except Exception:
        raise Exception(
            f"Access token creation failed. Reponse from the server was: {r.json()}"
        )

    token = r.json()

    return token


def most_recent(product_name, latlim, lonlim, collection_name):
    product_type = product_name.split("_")[0]

    base_url = "https://catalogue.dataspace.copernicus.eu/odata/v1/Products"
    filters = [
        f"Attributes/OData.CSC.StringAttribute/any(att:att/Name eq 'productType' and att/OData.CSC.StringAttribute/Value eq '{product_type}')",
        f"OData.CSC.Intersects(area=geography'SRID=4326;POLYGON(({lonlim[0]} {latlim[0]},{lonlim[1]} {latlim[0]},{lonlim[1]} {latlim[1]},{lonlim[0]} {latlim[1]},{lonlim[0]} {latlim[0]}))')",
        f"Collection/Name eq '{collection_name}'",
    ]

    query = base_url + "?$filter=" + " and ".join(filters)

    query += "&$orderby=ContentDate/Start desc"

    results = {"@odata.nextLink": query}

    scenes = list()
    out = requests.get(results["@odata.nextLink"])
    out.raise_for_status()
    results = out.json()
    scenes += results["value"]

    if len(scenes) == 0:
        return None
    else:
        return pd.Timestamp(scenes[-1]["ContentDate"]["Start"]).to_pydatetime()


def download(
    folder, latlim, lonlim, timelim, product_name, product_type, node_filter=None
):
    sd = datetime.datetime.strftime(timelim[0], "%Y-%m-%dT00:00:00Z")
    ed = datetime.datetime.strftime(timelim[1], "%Y-%m-%dT23:59:59Z")

    # NOTE paths on windows have a max length, this extends the max length, see
    # here for more info https://learn.microsoft.com/en-us/windows/win32/fileio/maximum-file-path-limitation?tabs=registry
    if os.name == "nt":
        cachedir = "\\\\?\\" + os.path.join(os.path.abspath(folder), "cache")
    else:
        cachedir = os.path.join(folder, "cache")

    memory = Memory(cachedir, verbose=0)

    product_name_ = {
        "SENTINEL3": "SENTINEL-3",
        "SENTINEL2": "SENTINEL-2",
        "SENTINEL-2": "SENTINEL-2",
        "SENTINEL-3": "SENTINEL-3",
    }[product_name]

    base_url = "https://catalogue.dataspace.copernicus.eu/odata/v1/Products"
    filters = [
        f"Attributes/OData.CSC.StringAttribute/any(att:att/Name eq 'productType' and att/OData.CSC.StringAttribute/Value eq '{product_type}')",
        f"OData.CSC.Intersects(area=geography'SRID=4326;POLYGON(({lonlim[0]} {latlim[0]},{lonlim[1]} {latlim[0]},{lonlim[1]} {latlim[1]},{lonlim[0]} {latlim[1]},{lonlim[0]} {latlim[0]}))')",
        f"Collection/Name eq '{product_name_}'",
        f"ContentDate/Start gt {sd}",
        f"ContentDate/Start lt {ed}",
    ]

    query = base_url + "?$filter=" + " and ".join(filters)

    results = {"@odata.nextLink": query}

    @memory.cache()
    def query_results(results):
        scenes = list()
        while "@odata.nextLink" in results.keys():
            out = requests.get(results["@odata.nextLink"])
            out.raise_for_status()
            results = out.json()
            scenes += results["value"]
        return scenes

    scenes = query_results(results)

    # Drop identical scenes.
    scene_names = {x["Name"]: i for i, x in enumerate(scenes)}
    scenes = [scenes[i] for i in scene_names.values()]

    if product_name == "SENTINEL-2":
        # Filter reprocessed scenes (select newest).
        overview = dict()
        for i, scene in enumerate(scenes):
            mission_id, level, stime, baseline, rel_orbit, tile_number = scene[
                "Name"
            ].split("_")[:6]
            key = "_".join([mission_id, level, stime, rel_orbit, tile_number])
            baseline_int = int(baseline[1:])
            if key not in list(overview.keys()):
                overview[key] = (baseline_int, scene)
            else:
                if baseline_int > overview[key][0]:
                    overview[key] = (baseline_int, scene)

        scenes = [x[1] for x in overview.values()]

    @memory.cache()
    def get_scene_nodes(scene_id):
        files = list()
        nodes_to_crawl = [
            f"https://catalogue.dataspace.copernicus.eu/odata/v1/Products({scene_id})/Nodes"
        ]

        while len(nodes_to_crawl) > 0:
            node_path = nodes_to_crawl[0]

            node_ = requests.get(node_path)
            node_.raise_for_status()
            node = node_.json()["result"]

            for sub_node in node:
                sub_node_name = sub_node["Name"]
                if sub_node["ChildrenNumber"] != 0:
                    nodes_to_crawl.append(node_path + f"({sub_node_name})/Nodes")
                else:
                    uri = sub_node["Nodes"]["uri"]
                    files.append(uri[:-6] if uri[-6:] == "/Nodes" else uri)

            nodes_to_crawl.remove(node_path)

        return files

    log.info(
        f"--> Searching nodes for {len(scenes)} `{product_name_}.{product_type}` scenes."
    )

    urls = list()
    fps = list()
    fp_checks = list()

    for scene in tqdm.tqdm(scenes, leave=False):
        nodes = get_scene_nodes(scene["Id"])
        if not isinstance(node_filter, type(None)):
            filtered_nodes = set([x for x in nodes if node_filter(x)])
        else:
            filtered_nodes = set(nodes)

        for node in filtered_nodes:
            node_path_rel = re.findall(r"Nodes\((.*?)\)", node)
            fp = os.path.join(folder, *node_path_rel)
            url = node + "/$value"
            fp_checks.append(fp)
            if not os.path.isfile(fp):
                urls.append(url)
                fps.append(fp)

    log.info(f"--> {len(fp_checks)} nodes required.")
    log.info(f"--> Downloading {len(urls)} missing nodes.").add()

    block_size = 10
    dled_fps = list()
    for i in range(0, len(urls), block_size):
        token = get_access_token()
        access_token = token["access_token"]
        headers = {"Authorization": f"Bearer {access_token}"}
        fps_ = download_urls(
            urls[i : i + block_size], None, fps=fps[i : i + block_size], headers=headers
        )
        dled_fps += fps_
        time.sleep(30)

    log.sub()

    for fp_ in fp_checks:
        if not os.path.isfile(fp_):
            log.warning(f"--> `{fp}` missing.")

    if "3" in product_name:
        dled_scenes = sorted(
            list(set([re.compile(r".*\.SEN3").search(x).group() for x in fp_checks]))
        )
    if "2" in product_name:
        dled_scenes = sorted(
            list(set([re.compile(r".*\.SAFE").search(x).group() for x in fp_checks]))
        )

    return list(dled_scenes)


def process_sentinel(
    scenes, variables, time_func, final_fn, post_processors, processor, bb=None
):
    """Process downloaded Sentinel scenes into netCDFs.

    Parameters
    ----------
    scenes : list
        Paths to downloaded nodes.
    variables : dict
        Keys are variable names, values are additional settings.
    time_func : function
        Function that parses a np.datetime64 from a filename.
    final_fn : str
        Path to the file in which to store all the combined data.
    post_processors : dict
        Functions to apply when the data has been processed.
    processor : function
        Function to apply sensor specific transformations.
    bb : list, optional
        Boundingbox to clip to, [xmin, ymin, xmax, ymax], by default None.

    Returns
    -------
    xr.Dataset
        Ouput data.

    Raises
    ------
    ValueError
        Invalid value for `source_name`.
    """

    chunks = {"time": 1, "x": 1000, "y": 1000}

    example_ds = None
    dss1 = dict()

    log.info(f"--> Processing {len(scenes)} scenes.").add()

    target_crs = rasterio.crs.CRS.from_epsg(4326)

    for i, scene_folder in enumerate(scenes):
        folder, fn = os.path.split(scene_folder)

        ext = os.path.splitext(scene_folder)[-1]

        fp = os.path.join(folder, os.path.splitext(fn)[0] + ".nc")
        if os.path.isfile(fp):
            ds = open_ds(fp)
            dtime = ds.time.values[0]
            if dtime in dss1.keys():
                dss1[dtime].append(ds)
            else:
                dss1[dtime] = [ds]
            continue

        if ext == ".zip":
            scene_folder = unpack(fn, folder)
            remove_folder = True
        else:
            scene_folder = scene_folder
            remove_folder = True

        ds, to_remove = processor(scene_folder, variables, bb=bb)

        # Apply variable specific functions.
        for vars in variables.values():
            for func in vars[2]:
                ds, label = apply_enhancer(ds, vars[1], func)

        # NOTE: see https://github.com/corteva/rioxarray/issues/545
        # NOTE: see rioxarray issue here: https://github.com/corteva/rioxarray/issues/570
        ds = ds.sortby("y", ascending=False)
        _ = [ds[var].rio.write_nodata(np.nan, inplace=True) for var in ds.data_vars]

        # Clip and pad to bounding-box
        if isinstance(example_ds, type(None)):
            example_ds = make_example_ds(ds, folder, target_crs, bb=bb)
        ds = ds.rio.reproject_match(example_ds).chunk(
            {k: v for k, v in chunks.items() if k in ["x", "y"]}
        )

        # NOTE: see rioxarray issue here: https://github.com/corteva/rioxarray/issues/570
        _ = [
            ds[var].attrs.pop("_FillValue")
            for var in ds.data_vars
            if "_FillValue" in ds[var].attrs.keys()
        ]

        ds = ds.assign_coords({"x": example_ds.x, "y": example_ds.y})

        dtime = time_func(fn)
        ds = ds.expand_dims({"time": 1})
        ds = ds.assign_coords({"time": [dtime]})
        ds.attrs = {}

        # Save to netcdf
        ds = save_ds(
            ds,
            fp,
            chunks=chunks,
            encoding="initiate",
            label=f"({i + 1}/{len(scenes)}) Processing {fn} to netCDF.",
        )

        if dtime in dss1.keys():
            dss1[dtime].append(ds)
        else:
            dss1[dtime] = [ds]

        for x in to_remove:
            remove_ds(x)

        rmve = {"NO": False, "YES": True}.get(
            os.environ.get("PYWAPOR_REMOVE_TEMP_FILES", "YES"), True
        )
        if remove_folder and rmve:
            if os.path.isdir(scene_folder):
                try:
                    shutil.rmtree(scene_folder)
                except PermissionError:
                    log.info(f"--> Unable to delete folder `{scene_folder}`.")

    log.sub()

    # Merge spatially.
    dss = [xr.concat(dss0, "stacked").median("stacked") for dss0 in dss1.values()]

    # Merge temporally.
    ds = xr.concat(dss, "time")

    # Define output path.
    fp = os.path.join(folder, final_fn)

    # Apply general product functions.
    ds = apply_enhancers(post_processors, ds)

    # Remove unrequested variables.
    ds = ds[list(post_processors.keys())]

    for var in ds.data_vars:
        ds[var].attrs = {}

    ds = ds.rio.write_crs(target_crs)

    ds = ds.sortby("time")

    # Save final netcdf.
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="All-NaN slice encountered")
        warnings.filterwarnings(
            "ignore", message="invalid value encountered in true_divide"
        )
        warnings.filterwarnings(
            "ignore", message="divide by zero encountered in true_divide"
        )
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        ds = save_ds(ds, fp, chunks=chunks, encoding="initiate", label="Merging files.")

    # Remove intermediate files.
    for dss0 in dss1.values():
        for x in dss0:
            remove_ds(x)

    if not isinstance(example_ds, type(None)):
        remove_ds(example_ds)

    return ds


if __name__ == "__main__":
    # product_name = "SENTINEL-2"
    # product_type = "S2MSI2A"

    product_name = "SENTINEL-3"
    product_type = "SL_2_LST___"

    timelim = [datetime.date(2023, 3, 1), datetime.date(2023, 3, 3)]
    latlim = [29.4, 29.7]
    lonlim = [30.7, 31.0]
    folder = r"/Users/hmcoerver/Local/new_sentinel_test"

    adjust_logger(True, folder, "INFO")

    # variables = {
    #     "_B01_60m.jp2": [(), "coastal_aerosol", []],
    #     "_B02_60m.jp2": [(), "blue", []],
    #     "_B03_60m.jp2": [(), "green", []],
    #     "_B04_60m.jp2": [(), "red", []],
    # }

    variables = {
        "LST_in.nc": [(), "lst", []],
        "geodetic_in.nc": [(), "coords", []],
    }

    def node_filter(node_info):
        fn = os.path.split(node_info)[-1]
        to_dl = list(variables.keys()) + ["MTD_MSIL2A.xml"]
        return np.any([x in fn for x in to_dl])

    # dled_scenes = download(folder, latlim, lonlim, timelim, product_name, product_type, node_filter = node_filter)

    s3_scene_name = "S3A_SL_2_LST____20230303T194903_20230303T195203_20230305T053809_0179_096_128_0360_PS1_O_NT_004.SEN3"

    s2_scene_name = "S2B_MSIL2A_20230302T083759_N0509_R064_T36RTT_20230302T132852.SAFE"



import copy
import os
from functools import partial

from pywapor.collect.protocol import cog
from pywapor.enhancers import lulc
from pywapor.general.processing_functions import open_ds


def default_vars(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary
    with metadata on which exact layers need to be requested from the server, how they should
    be renamed, and how their dimensions are defined.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Metadata on which exact layers need to be requested from the server.
    """
    variables = {
        "2009_V2.3_Global": {
            "Band1": [("lat", "lon"), "lulc"],
            "crs": [(), "spatial_ref"],
        },
    }

    req_dl_vars = {
        "2009_V2.3_Global": {
            "lulc": ["Band1", "crs"],
            "rs_min": ["Band1", "crs"],
            "z_obst_max": ["Band1", "crs"],
            "land_mask": ["Band1", "crs"],
            "lue_max": ["Band1", "crs"],
        },
    }

    out = {
        val: variables[product_name][val]
        for sublist in map(req_dl_vars[product_name].get, req_vars)
        for val in sublist
    }

    return out


def remove_var(ds, var):
    return ds.drop_vars([var])


def default_post_processors(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary with a
    list of functions per variable that should be applied after having collected the data
    from a server.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Functions per variable that should be applied to the variable.
    """

    post_processors = {
        "2009_V2.3_Global": {
            "lulc": [],
            "rs_min": [
                partial(
                    lulc.lulc_to_x,
                    in_var="lulc",
                    out_var="rs_min",
                    convertor=lulc.globcover_to_rs_min(),
                )
            ],
            "z_obst_max": [
                partial(
                    lulc.lulc_to_x,
                    in_var="lulc",
                    out_var="z_obst_max",
                    convertor=lulc.globcover_to_z_obst_max(),
                )
            ],
            "land_mask": [
                partial(
                    lulc.lulc_to_x,
                    in_var="lulc",
                    out_var="land_mask",
                    convertor=lulc.globcover_to_land_mask(),
                )
            ],
            "lue_max": [
                partial(
                    lulc.lulc_to_x,
                    in_var="lulc",
                    out_var="lue_max",
                    convertor=lulc.globcover_to_lue_max(),
                )
            ],
        },
    }

    out = {k: v for k, v in post_processors[product_name].items() if k in req_vars}

    return out


def url_func(product_name):
    """Returns a url at which to collect GLOBCOVER data.

    Parameters
    ----------
    product_name : None
        Not used.

    Returns
    -------
    str
        The url.
    """
    return r"http://due.esrin.esa.int/files/GLOBCOVER_L4_200901_200912_V2.3.color.tif"


def most_recent(product_name, *args):
    return None


def download(
    folder,
    latlim,
    lonlim,
    product_name,
    req_vars=["lulc"],
    variables=None,
    post_processors=None,
    **kwargs,
):
    """Download GLOBCOVER data and store it in a single netCDF file.

    Parameters
    ----------
    folder : str
        Path to folder in which to store results.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    product_name : str, optional
        Name of the product to download, by default "P05".
    req_vars : list, optional
        Which variables to download for the selected product, by default ["p"].
    variables : dict, optional
        Metadata on which exact layers need to be requested from the server, by default None.
    post_processors : dict, optional
        Functions per variable that should be applied to the variable, by default None.

    Returns
    -------
    xr.Dataset
        Downloaded data.
    """

    folder = os.path.join(folder, "GLOBCOVER")

    fn = os.path.join(folder, f"{product_name}.nc")
    req_vars_orig = copy.deepcopy(req_vars)
    if os.path.isfile(fn):
        existing_ds = open_ds(fn)
        req_vars_new = list(set(req_vars).difference(set(existing_ds.data_vars)))
        if len(req_vars_new) > 0:
            req_vars = req_vars_new
            existing_ds = existing_ds.close()
        else:
            return existing_ds[req_vars_orig]

    spatial_buffer = True
    if spatial_buffer:
        latlim = [latlim[0] - 0.01, latlim[1] + 0.01]
        lonlim = [lonlim[0] - 0.01, lonlim[1] + 0.01]

    coords = {"x": ("lon", lonlim), "y": ("lat", latlim)}

    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {
            k: {True: default_processors[k], False: v}[v == "default"]
            for k, v in post_processors.items()
            if k in req_vars
        }

    ds = cog.download(
        fn, product_name, coords, variables, post_processors, url_func, ndv=0
    )

    return ds[req_vars_orig]


if __name__ == "__main__":
    product_name = "2009_V2.3_Global"

    folder = r"/Users/hmcoerver/Local/globcov_test"
    latlim = [28.9, 29.7]
    lonlim = [30.2, 31.2]
    variables = None
    post_processors = None
    req_vars = ["lulc", "rs_min", "land_mask"]

    ds = download(
        folder,
        latlim,
        lonlim,
        product_name,
        req_vars=req_vars,
        variables=variables,
        post_processors=post_processors,
    )



def drop_empty_times(ds, x, drop_vars = ["lst"], *args):
    """Checks for time coordinates at which there is no valid to for any of the variables in
    `drop_vars` and removes those coordinates.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to check.
    x : None
        Not used.
    drop_vars : list, optional
        Variables to check for valid data, by default ["lst"].

    Returns
    -------
    xr.Dataset
        Dataset with adjusted time coordinates.
    """
    for drop_var in drop_vars:
        ds = ds.isel(time = (ds[drop_var].count(dim=["y", "x"]) > 0).values)
    return ds